/*-------------------------------------------------------------------
  Authors: 2001: Stijn Vansummeren

  Test program for the BTree project
  -------------------------------------------------------------------*/

import java.io.*;
import java.util.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.catalog.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.file.*;
import be.ulb.db.index.*;
import be.ulb.db.schema.*;

public class TestLoad
{
    public static final int BLOCKSIZE = 512;
    public static final int NUMBLOCKS = 2500;
    public static final int NUMBUFFERS = 100;

    //number of records to add
    public static final int NUMRECORDS = 300; 

    //Rationale of NUMRECORDS:
    //----------------------- 
    //After inserting NUMRECORDS into the Index, the BTree will have
    //at least three levels.  This is the minimum number of levels we
    //need in order to test if all block splitting and merging works
    //correctly.
    //
    //Why will it be three levels? To store a (keyvalue, recordaddr)
    //value in a BTree index leaf block, for the specific schema of
    //keyvalues used below, we need at least 32 bytes (unicode
    //characters take two bytes). Hence, a leaf block can store at
    //most 16 (keyvalue, recordaddr) pairs. An internal BT index block
    //stores (keyvalue, blockaddr) pairs (since we don't need a
    //complete recordaddress for internal BT index blocks), which
    //takes a minimum of 28 bytes. An internal BT index block can
    //hence store 18 (keyvalue, blockaddr) pairs. A two-level BTree
    //can hence store at most 18 * 16 = 288 recors. Hence, with a
    //blocksize of 512 bytes we should reach a three-level BTree index
    //after insertion of 289 records (we take 300 to be sure).

    public static void printStage(String stage)
    {
	System.out.println("");
	System.out.println("*********************************************");
	System.out.println("Stage: " + stage + " ...");
	System.out.println("*********************************************");  
    }

    public static void main(String [] args)
    {
	try {
	    
	    File file = new File("oef.db");
	    if (file.exists())
		file.delete();

	    printStage("Creating database");
	
	    //create a database with NUMBLOCK blocks of BLOCKSIZE bytes, and a
	    //buffermanager with NUMBUFFERS buffers
	    Database db = new Database(file, NUMBLOCKS, BLOCKSIZE, NUMBUFFERS);

	    String [] names = { "A", "B", "C" };
	    AttributeType [] types = { new AttributeTypeString(10), 
				       new AttributeTypeInt(), 
				       new AttributeTypeBlockAddress() };
	    Schema schema = new Schema(names, types);

	    String [] keynames = { "B", "A"};
	    Schema keySchema = schema.project(keynames);

	    printStage("Creating index");
	    Index index = new BTIndex(keySchema, db.getBufferManager());
	
	    printStage("Adding tuples to index");
	    Record keyvalue = new Record(keySchema);
	    RecordAddress addr;
	    
	    for (int i = 0; i < NUMRECORDS; i++) {
		keyvalue.putInt("B", new Integer(i));
		keyvalue.putString("A", "String " + i);
		//create a fake record address
		addr = new RecordAddress(new BlockAddress(i), i);
		index.addRecord(keyvalue, addr);
	    }

	    printStage("Verifying");
	    boolean globalok = true;
	    for (int i = 0; i < NUMRECORDS; i++) {
		keyvalue.putInt("B", new Integer(i));
		keyvalue.putString("A", "String " + i);
		RecordAddressIterator it = index.findRecord(keyvalue);
		boolean ok = false;
		if (it.hasNext()) {
		    addr = it.next();
		    ok = (addr.getRecordIndex() == i) && (addr.getBlockAddress().getBlockOffset() == i);
		}
		globalok &= ok;
		if (!ok)
		    System.out.println("Error: searchkey " + i + "not found");
	    }
	    if (globalok)
		System.out.println("Sanity check passed");

	    printStage("Removing tuples");
	    //delete all even tuples - the btree should become a two-level btree
	    for (int i = 0; i < NUMRECORDS; i+=2) {
		keyvalue.putInt("B", new Integer(i));
		keyvalue.putString("A", "String " + i);
		addr = new RecordAddress(new BlockAddress(i), i);
		index.removeRecord(keyvalue, addr);
	    }
	    
	    printStage("Verifying");
	    for (int i = 0; i < NUMRECORDS; i++) {
		keyvalue.putInt("B", new Integer(i));
		keyvalue.putString("A", "String " + i);
		RecordAddressIterator it = index.findRecord(keyvalue);
		boolean ok = (i % 2 == 0);
		if (it.hasNext()) {
		    addr = it.next();
		    ok = (addr.getRecordIndex() == i) && (addr.getBlockAddress().getBlockOffset() == i) && (i % 2 != 0);
		}
		if (!ok && (i % 2 == 0))
		    System.out.println("Error: searchkey " + i + "found, where it should not be");

		if (!ok && (i % 2 != 0))
		    System.out.println("Error: searchkey " + i + "not found");

		globalok &= ok;
	    }
	    if (globalok)
		System.out.println("Sanity check passed");
	    
	}
	catch(Exception e) {
	    System.out.println("Exception caught!");
	    e.printStackTrace();
	}
    }
}
