/*-------------------------------------------------------------------
   Authors: 2001-2003: Stijn Vansummeren

   The Block class is a base class wich wraps memory buffers holding
   information of a block in secondary storage. Subclasses of the
   Block class can provide extra features to correctly handle a
   specific kind of block.For instance, the MainBlock subclass
   provides features to insert, update and modify records in the
   wrapped buffer. We can envision subclasses handling index blocks as
   well.

   $Id: Block.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db;

import java.io.*;
import java.nio.*;
import java.lang.*;
import java.util.*;

public abstract class Block
{
    /* The in-memory buffer in wich we store the information of the
     * block. */
    protected ByteBuffer buff;

    /* It's address */
    protected BlockAddress addr;

    /* Wrap around the buffer buff, and note that we have block
     * address addr */
    public Block(ByteBuffer buff, BlockAddress addr)
    {
	this.buff = buff;
	this.addr = addr;
	if (buff.isReadOnly()) {
	    System.out.println("Block " + addr.getBlockOffset() + " wrapped around read-only buffer!");
	}
    }

    /* This method is used by the buffer manager to let a block
     * initialize it's buffer when the block is wrapped around a newly
     * allocated buffer. This can come in handy for instance when we
     * need to initialize a header of some kind. In principle there
     * can be "garbage data" in a newly allocated buffer (from the
     * previous allocation). By initializing such buffers we make sure
     * that this data is not falsly interpreted in the new
     * block. Subclasses should always implement this method. */
    public abstract void initContents();

    /* This method is used by the buffer manager to notify a block
     * that it is being detached from its buffer (because the block is
     * unpinned or freed). Subclasses can override this method to
     * perform some cleanup if necessary. */
    public void detach()
    {
	/* make sure that a block which is detached from its buffer
	 * cannot write to the buffer anymore */
	this.buff = null;
	this.addr = null;
    }

    /* returns the address of this block */
    public BlockAddress getBlockAddress()
    {
	return addr;
    }

    /* returns a read-only version of the wrapped buffer */
    public final ByteBuffer getByteBuffer()
    {
	buff.position(0);
	buff.limit(buff.capacity());
	
	return buff.asReadOnlyBuffer();
    }

    /* subclasses should return true if the modifications they have
       made to the wrapped buffer has to be saved back to disk */
    public abstract boolean isModified();
}

