 /*-------------------------------------------------------------------
   Authors: 2001-2003: Stijn Vansummeren

   A record is a collection of attributes, adhering to some schema.
   The information contained in the record is kept as a sequence of
   bytes in a ByteBuffer.

   $Id: Record.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db;

import java.nio.*;

import be.ulb.db.schema.*;

public class Record implements Comparable<Record>
{
    private Schema schema; /* describes the schema of this record */
    private ByteBuffer buff;

    /* Create a new record with its own buffer adhering to schema */
    public Record(Schema schema)
    {
	this.schema = schema;
	this.buff = ByteBuffer.allocate(schema.bytesize());
    }

    /* Create a new record adhering to schema, wrapping around the
     * buffer buff.  */
    public Record(Schema schema, ByteBuffer buff) throws SchemaException
    {
	this.schema = schema;
	this.buff = buff;
	if (this.buff.capacity() != schema.bytesize()) 
	    throw new SchemaException("The size of the buffer does not agree with the size of the schema!"
				      + buff.capacity() + "<>" + schema.bytesize());
    }

    /* return the schema this record belongs to */
    public Schema getSchema()
    {
	return schema;
    }

    /* returns a read-only version of the buffer we use to store the
     * record information  */
    public ByteBuffer getByteBuffer()
    {
	buff.position(0);
	buff.limit(buff.capacity());
	
	return buff.asReadOnlyBuffer();
    }


    /* returns a new record which projects this record according to
     * projectSchema. Throws SchemaException if projectSchema is not a
     * subschema of the record schema. TODO: write more efficient? */
    public Record project(Schema projectSchema) throws SchemaException
    {
	if (!projectSchema.isSubSchemaOf(schema))
	    throw new SchemaException("The schema to be projected upon is not a subschema of the record's schema!");
	
	Record rec = new Record(projectSchema);
	for (int i = 0; i < projectSchema.length(); i++) 
	    rec.put(i, this.get(projectSchema.getName(i)));
	    
	return rec;
    }

    /* Put the content into the attribute with attributeindex
     * index. Throws SchemaException if the type of that
     * attribute is incompatible with the provided content. Throws
     * SchemaException if the index is not in the schema.
     */
    public void put(int index, Object content) throws SchemaException
    {
	AttributeType ty = schema.getType(index);

	if (!ty.hasInstance(content)) 
	    throw new SchemaException("The value to be put in attribute " + index +
					    " is incompatible with the type given in the schema.");
	
	/* Splice off a new subbuffer which has the position and
	   limit set correctly.  This will make sure that
	   writeInstanceToBuffer does not write more than is
	   allowed */
	int startpos = schema.getOffset(index); 
	int endpos = startpos + ty.bytesize();

	/* position can never be greater than the limit - therefore we first set the limit */
	buff.limit(endpos); 
	buff.position(startpos);
	ByteBuffer subbuff = buff.slice();

	ty.writeInstanceToByteBuffer(subbuff, content);
    }

    /* Put the content into the attribute with attributename name.
     * Throws SchemaException if the type of that attribute is
     * incompatible with the provided content. */
    public void put(String name, Object content) throws SchemaException
    {
	put(schema.resolveName(name), content);
    }

    /* Read the content of attributeindex index as a Java object. Throws
     * SchemaException if the attributeindex is not in the schema. */
    public Object get(int index) throws SchemaException
    {
	AttributeType ty = schema.getType(index);

	/* Splice off a new subbuffer which hase the position and limit set correctly.
	   This will make sure that writeInstanceToBuffer does not write more than is allowed */
	int startpos = schema.getOffset(index); 
	int endpos = startpos + ty.bytesize();

	/* position can never be greater than the limit - therefore we first set the limit */
	buff.limit(endpos);
	buff.position(startpos);
	ByteBuffer subbuff = buff.slice();

	return ty.readInstanceFromByteBuffer(subbuff);
    }

    /* Read the content of attribute with attributename name as a Java
     * object */
    public Object get(String name) throws SchemaException
    {
	return get(schema.resolveName(name));
    }

    /* put a string in attribute index. Throws SchemaException
     * if that attribute does not have a string type. Throws
     * SchemaEception if the attributeindex is not in the schema. */
    public void putString(int index, String content) throws SchemaException
    {
	put(index, content);
    }

    /* put a string in attribute with attributename name. Throws
     * SchemaException if that attribute does not have a string
     * type. Throws SchemaEception if the attribute is not in the
     * schema. */
    public void putString(String name, String content) throws SchemaException
    {
	putString(schema.resolveName(name), content);
    }
    

    /* Read the content of attribute with attributeindex index as a
     * String. Throws SchemaException if that attribute does not
     * have a string type. Throws SchemaException if the attribute is
     * not in the schema. */
    public String getString(int index) throws SchemaException
    {
	if (schema.getType(index) instanceof AttributeTypeString)
	    return (String) get(index);
	else
	    throw new SchemaException("Attribute " + index + " does not have a string type.");
    }

    /* Read the content of attribute with attributename name as a
     * String. Throws SchemaException if that attribute does not
     * have a string type. Throws SchemaException if the attribute is
     * not in the schema. */
    public String getString(String name) throws SchemaException
    {
	return getString(schema.resolveName(name));
    }
    
    /* put an integer in attribute with number index. Throws
     * SchemaException if that attribute does not have a string
     * type. Throws SchemaException if the attribute is not in the
     * schema. */
    public void putInt(int index, Integer content) throws SchemaException
    {
	put(index, content);
    }

    /* put an integer in attribute with attributename name. Throws
     * SchemaException if that attribute does not have a string
     * type. Throws SchemaException if the attribute is not in the
     * schema. */
    public void putInt(String name, Integer content) throws SchemaException
    {
	putInt(schema.resolveName(name), content);
    }


    /* Read the content of attribute with number index as an
     * integer. Throws SchemaException if that attribute does
     * not have a integer type. Throws SchemaException if the
     * attribute is not in the schema. */
    public Integer getInt(int index) throws SchemaException
    {
	if (schema.getType(index) instanceof AttributeTypeInt)
	    return (Integer) get(index);
	else
	    throw new SchemaException("Attribute " + index + " does not have an integer type.");
    }

    /* Read the content of attribute with attributename name as an
     * integer. Throws SchemaException if that attribute does
     * not have a integer type. Throws SchemaException if the
     * attribute is not in the schema. */
    public Integer getInt(String name) throws SchemaException
    {
	return getInt(schema.resolveName(name));
    }

    /* put a BlockAddress in attribute with number index. Throws
     * SchemaException if that attribute does not have a
     * BlockAddress type. Throws SchemaException if the attribute is
     * not in the schema. */
    public void putBlockAddress(int index, BlockAddress content) throws SchemaException
    {
	put(index, content);
    }

    /* put an BlockAddress in attribute with attributename
     * name. Throws SchemaException if that attribute does not
     * have a BlockAddress type. Throws SchemaException if the
     * attribute is not in the schema. */
    public void putBlockAddress(String name, BlockAddress content) throws SchemaException
    {
	putBlockAddress(schema.resolveName(name), content);
    }

    /* Read the content of attribute with number index as an
     * BlockAddress. Throws SchemaException if that attribute does
     * not have a BlockAddress type. Throws SchemaException if the
     * attribute is not in the schema. */
    public BlockAddress getBlockAddress(int index) throws SchemaException
    {
	if (schema.getType(index) instanceof AttributeTypeBlockAddress)
	    return (BlockAddress) get(index);
	else
	    throw new SchemaException("Attribute " + index + " does not have a BlockAddress type.");
    }

    /* Read the content of attribute with attributename name as an
     * BlockAddress. Throws SchemaException if that attribute does
     * not have a BlockAddress type. Throws SchemaException if the
     * attribute is not in the schema. */
    public BlockAddress getBlockAddress(String name) throws SchemaException
    {
	return getBlockAddress(schema.resolveName(name));
    }

    /* put a RecordAddress in attribute with number index. Throws
     * SchemaException if that attribute does not have a
     * RecordAddress type. Throws SchemaException if the attribute is
     * not in the schema. */
    public void putRecordAddress(int index, RecordAddress content) throws SchemaException
    {
	put(index, content);
    }

    /* put an RecordAddress in attribute with attributename
     * name. Throws SchemaException if that attribute does not
     * have a RecordAddress type. Throws SchemaException if the
     * attribute is not in the schema. */
    public void putRecordAddress(String name, RecordAddress content) throws SchemaException
    {
	putRecordAddress(schema.resolveName(name), content);
    }

    /* Read the content of attribute with number index as an
     * RecordAddress. Throws SchemaException if that attribute does
     * not have a RecordAddress type. Throws SchemaException if the
     * attribute is not in the schema. */
    public RecordAddress getRecordAddress(int index) throws SchemaException
    {
	if (schema.getType(index) instanceof AttributeTypeRecordAddress)
	    return (RecordAddress) get(index);
	else
	    throw new SchemaException("Attribute " + index + " does not have a RecordAddress type.");
    }

    /* Read the content of attribute with attributename name as an
     * RecordAddress. Throws SchemaException if that attribute does
     * not have a RecordAddress type. Throws SchemaException if the
     * attribute is not in the schema. */
    public RecordAddress getRecordAddress(String name) throws SchemaException
    {
	return getRecordAddress(schema.resolveName(name));
    }


    /* put a Boolean in attribute with number index. Throws
     * SchemaException if that attribute does not have a
     * Boolean type. Throws SchemaException if the attribute is
     * not in the schema. */
    public void putBoolean(int index, Boolean content) throws SchemaException
    {
	put(index, content);
    }

    /* put an Boolean in attribute with attributename
     * name. Throws SchemaException if that attribute does not
     * have a Boolean type. Throws SchemaException if the
     * attribute is not in the schema. */
    public void putBoolean(String name, Boolean content) throws SchemaException
    {
	putBoolean(schema.resolveName(name), content);
    }

    /* Read the content of attribute with number index as an
     * Boolean. Throws SchemaException if that attribute does
     * not have a Boolean type. Throws SchemaException if the
     * attribute is not in the schema. */
    public Boolean getBoolean(int index) throws SchemaException
    {
	if (schema.getType(index) instanceof AttributeTypeBoolean)
	    return (Boolean) get(index);
	else
	    throw new SchemaException("Attribute " + index + " does not have a Boolean type.");
    }

    /* Read the content of attribute with attributename name as an
     * Boolean. Throws SchemaException if that attribute does
     * not have a Boolean type. Throws SchemaException if the
     * attribute is not in the schema. */
    public Boolean getBoolean(String name) throws SchemaException
    {
	return getBoolean(schema.resolveName(name));
    }

    /* compare this record with another record of the same
     * schema. Returns <= 0 is this record is less than the other, 0
     * if it is equal, and >= 0 if it is greater. 
     * TODO: rewrite to obtain a more natural ordering.
     */ 
    public int compareTo(Record other)
    {
	if (!other.getSchema().equals(schema))
	    throw new IllegalArgumentException("Can only compare records with the same schema.");

	/* we compare lexicographically on the byte buffer */
	buff.position(0);
	buff.limit(buff.capacity());
	other.buff.position(0);
	other.buff.limit(other.buff.capacity());
	
	for (int i = 0; i < buff.capacity(); i++) {
	    byte b1 = buff.get(i);
	    byte b2 = other.buff.get(i);
	    
	    if (b1 < b2)
		return -1;
	    else if (b1 > b2)
		return 1;
	}

	return 0;
    }

    public boolean equals(Object o)
    {
	if (!(o instanceof Record))
	    return false;

	return compareTo((Record) o) == 0;
    }
}


