/*-------------------------------------------------------------------
   Authors: 2001-2003: Stijn Vansummeren

   The relation class represents relations. It contains a file to hold
   its data records, and has knowledge about the indexes on the data
   file. It uses the system catalog to store it relevant information
   (where the data file starts, the schema, the indexes, ...).

   $Id: Relation.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
   -------------------------------------------------------------------*/
package be.ulb.db;


import java.util.*;
import java.io.*;
import java.lang.ref.*;

import be.ulb.db.buffermanager.*;
import be.ulb.db.catalog.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.file.*;
import be.ulb.db.index.*;
import be.ulb.db.schema.*;

public class Relation
{
    protected BufferManager bm; //the associated buffer manager
    protected Schema schema; //the shema of the records in this relation
    protected Catalog catalog; //the system catalog
    protected String name; //the name of the relation
    protected MainFile mainfile; //the mainfile used to store data records

    protected int maxid;
    protected Hashtable schema_to_index; //mapping of key schema's to indexes
    protected Hashtable schema_to_id; //mapping of key schema's to the indexes ID

    /* we maintain weak references to our spawned iterators so that we
     * can close them when the file is closed */
    private List iterators;

    /** creates a new relation adhering to schema, with name name */
    public Relation(String name, Schema schema, Catalog catalog, BufferManager bm)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.name = name;
	this.schema = schema;
	this.catalog = catalog;
	this.bm = bm;
	this.iterators = new LinkedList();

	this.schema_to_index = new Hashtable();
	this.schema_to_id = new Hashtable();

	/* create a new main file */
	this.mainfile = new MainFile(bm, schema);
	
	/* add the information to the catalog */
	catalog.getRelationCatalog().add(name, mainfile.getBlockAddress(), schema);
	catalog.getAttributeCatalog().add(name, schema);
    }
    
    /** loads an existing relation */
    public Relation(String name, Catalog catalog, BufferManager bm)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.name = name;
	this.catalog = catalog;
	this.bm = bm;
	this.schema_to_index = new Hashtable();
	this.schema_to_id = new Hashtable();

	try {
	    Record relrec = catalog.getRelationCatalog().find(name);
	    if (relrec == null)
		throw new NoSuchElementException("Relation " + name + " not found.");

	    /* ----- LOAD THE MAIN FILE ----- */
	    BlockAddress addr = relrec.getBlockAddress(RelationCatalogSchema.addrIndex);
	    int attrnum = relrec.getInt(RelationCatalogSchema.attrNumIndex).intValue();
	    String [] names = new String[attrnum];
	    AttributeType [] types = new AttributeType[attrnum];

	    //reconstruct the schema
	    RecordIterator it = catalog.getAttributeCatalog().find(name);
	    int count = 0;
	    while (it.hasNext()) {
		Record schemrec = it.next();
		int offset = schemrec.getInt(AttributeCatalogSchema.offsetIndex).intValue();
		int length = schemrec.getInt(AttributeCatalogSchema.lengthIndex).intValue();
		String attrname = schemrec.getString(AttributeCatalogSchema.attrNameIndex);
		String typeclassname = schemrec.getString(AttributeCatalogSchema.typeIndex);
		
		names[offset] = attrname;
		types[offset] = AttributeTypeLoader.load(typeclassname, length);

		count++;
	    }

	    if (count != attrnum)
		throw new CatalogException("Catalog is corrupt - number of attributes stored for " + name + " is incorrect.");
	    
	    //create the mainfile object
	    schema = new Schema(names, types);
	    mainfile = new MainFile(bm, new Schema(names, types));

	    /* ----- LOAD THE INDEXES ----- */
	    it = catalog.getIndexCatalog().find(name);
	    while (it.hasNext()) {
		Record indexrec = it.next();
		int id = indexrec.getInt(IndexCatalogSchema.idIndex).intValue();
		attrnum = indexrec.getInt(IndexCatalogSchema.attrNumIndex).intValue();
		String indexclassname = indexrec.getString(IndexCatalogSchema.typeIndex);
		addr = indexrec.getBlockAddress(IndexCatalogSchema.addrIndex);

		names = new String[attrnum];
		types = new AttributeType[attrnum];
		
		//find out what the maximum id is up until now
		maxid = Math.max(maxid,id );

		//reconstruct the key's schema
		RecordIterator attit = catalog.getIndexAttrCatalog().find(name, id);
		count = 0;
		while (attit.hasNext()) {
		    Record indexattrrec = attit.next();
		    String attrname = indexattrrec.getString(IndexAttrCatalogSchema.attrNameIndex);
		    int offset = indexattrrec.getInt(IndexAttrCatalogSchema.offsetIndex).intValue();
		    Record schemrec = catalog.getAttributeCatalog().find(name, attrname);
		    if (schemrec == null)
			throw new CatalogException("Catalog is corrupt -  attribute stored for index " +
						   id + "in relation " + name + " not found.");

		    int length = schemrec.getInt(AttributeCatalogSchema.lengthIndex).intValue();
		    String typeclassname = schemrec.getString(AttributeCatalogSchema.typeIndex);
		
		    names[offset] = attrname;
		    types[offset] = AttributeTypeLoader.load(typeclassname, length);

		    count++;
		}

		if (count != attrnum)
		    throw new CatalogException("Catalog is corrupt - number of attributes stored for index "
					       + id + " on relation " + name + " is incorrect.");

		/* create the key's schema */
		Schema keyschema = new Schema(names, types);
		/* load the index */
		Index index = IndexLoader.load(indexclassname, keyschema, bm, addr);
		
		schema_to_index.put(keyschema, index);
		schema_to_id.put(keyschema, new Integer(id));
	    }
	    
	}
	catch (SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
    }

    //completely deletes the relation (including catalog information, data file and indexes)
    public void destroy()
	throws BufferManagerException, DiskManagerException, IOException
    {
	//destroy the indexes
	Enumeration e = schema_to_index.elements();
	while (e.hasMoreElements()) {
	    Index index = (Index) e.nextElement();
	    index.destroy();
	}

	//remove them from the catalog
	e = schema_to_id.elements();
	while (e.hasMoreElements()) {
	    int id = ((Integer) e.nextElement()).intValue();
	    catalog.getIndexCatalog().remove(this.name, id);
	    catalog.getIndexAttrCatalog().remove(this.name, id);
	}

	mainfile.destroy();

	catalog.getRelationCatalog().remove(this.name);
	catalog.getAttributeCatalog().remove(this.name);

	close();
    }

    //close the relation by freeing the associated resources. Subsequent operations may fail
    public void close()
    {
	mainfile.close();

	//close all indexes
	Enumeration e = schema_to_index.elements();
	while (e.hasMoreElements()) {
	    Index index = (Index) e.nextElement();
	    index.close();
	}
	schema_to_index.clear();

	//close all spawned iterators
	ListIterator it = iterators.listIterator();
	while (it.hasNext()) {
	    WeakReference ref = (WeakReference) it.next();
	    RecordIterator rit = (RecordIterator) ref.get();
	    if (rit != null)
		rit.close();
	}
	iterators.clear();
    }

    //Always close when we get garbage collected
    public void finalize()
    {
	close();
    }

    //returns the mainfile associated with this relation
    public MainFile getMainfFile()
    {
	return mainfile;
    }

    //returns the schema of this relation
    public Schema getSchema()
    {
	return schema;
    }
    
    /* creates a new index on keys having schema keySchema */
    public void createIndex(Schema keySchema, IndexFactory factory)
	throws BufferManagerException, DiskManagerException, IOException, SchemaException
    {
	Index index = factory.create(keySchema, bm);
		
	/* add all tuples in the relation to the index */
	RecordIterator it = mainfile.getIterator();
	while (it.hasNext()) {
	    Record rec = it.next();
	    Record keyval = rec.project(keySchema);
	    index.addRecord(keyval, it.getRecordAddress());
	}

	//add it to our map
	maxid++;
	schema_to_index.put(keySchema, index);
	schema_to_id.put(keySchema, new Integer(maxid));

	//free resources as soon as possible
	it.close(); 

	//store the index information in the catalog
	catalog.getIndexCatalog().add(this.name, index, maxid);
	catalog.getIndexAttrCatalog().add(this.name, index, maxid);
    }

    /* removes index on schema keySchema */
    public void removeIndex(Schema keySchema) 
	throws IOException, BufferManagerException, DiskManagerException, SchemaException
    {
	Index index = (Index) schema_to_index.get(keySchema);

	if (index == null) 
	    throw new NoSuchElementException("The relation has no index with the specified key schema.");
	
	//remove the index information from the catalog
	int id = ((Integer) schema_to_id.get(keySchema)).intValue();
	catalog.getIndexCatalog().remove(this.name, id);
	catalog.getIndexAttrCatalog().remove(this.name, id);

	//destroy the index
	index.destroy();
    }

    /* get the index on keyvalues adhering to keySchema */
    public Index getIndex(Schema keySchema)
    {
	return (Index) schema_to_index.get(keySchema);
    }
 
    /* Add a record to the mainfile AND update the indexes. */
    public void addRecord(Record record)
	throws IOException, BufferManagerException, DiskManagerException, SchemaException
    {
	RecordAddress addr = mainfile.addRecord(record);
	Enumeration e = schema_to_index.elements();
	
	while (e.hasMoreElements()) {
	    Index index = (Index) e.nextElement();
	    Record keyvalue = record.project(index.getSchema());

	    index.addRecord(keyvalue, addr);
	}
    }

    /* Updates the record in the mainfile with RecordAddress addr. Also
     * updates the index. */
    public void removeRecord(RecordAddress addr, Record rec) 
	throws IOException, BufferManagerException, DiskManagerException, SchemaException
    {
	Record oldrec = mainfile.getRecord(addr);

	//remove it, and put it back in again
	Enumeration e = schema_to_index.elements();
	while (e.hasMoreElements()) {
	    Index index = (Index) e.nextElement();
	    Record oldkeyvalue = oldrec.project(index.getSchema());
	    Record keyvalue = rec.project(index.getSchema());

	    index.removeRecord(oldkeyvalue, addr);
	    index.addRecord(keyvalue, addr);
	}

	mainfile.updateRecord(addr, rec);
    }


    /* Remove the record in the mainfile with RecordAddress addr. Also
     * removes it from the index. */
    public void removeRecord(RecordAddress addr) 
	throws IOException, BufferManagerException, DiskManagerException, SchemaException
    {
	Record rec = mainfile.getRecord(addr);
	
	Enumeration e = schema_to_index.elements();
	while (e.hasMoreElements()) {
	    Index index = (Index) e.nextElement();
	    Record keyvalue = rec.project(index.getSchema());

	    index.removeRecord(keyvalue, addr);
	}

	mainfile.removeRecord(addr);
    }


    /* Find the record(s) whose keyvalue attributes values match the
     * keyvalue values. Uses an available index if possible, if not a
     * subindex, and if none is available a direct scan on the
     * mainfile */
    public RecordIterator findRecord(Record keyvalue) 
	throws IOException, BufferManagerException, DiskManagerException, SchemaException
    {
	RecordIterator it = null;
	
	Index index = (Index) schema_to_index.get(keyvalue.getSchema());
	if (index != null) {
	    return new IndexIterator(mainfile, index, keyvalue);
	}
	else {
	    Enumeration e = schema_to_index.elements();
	    boolean found = false;
	    while (!found && e.hasMoreElements()) {
		index = (Index) e.nextElement();
		found = index.getSchema().isSubSchemaOf(keyvalue.getSchema());
	    }
	    
	    if (found) {
		//these iterators we have to close when the relation is closed
		it = new SubIndexIterator(mainfile, index, keyvalue);
		WeakReference ref = new WeakReference(it);
		iterators.add(ref);
	    }
	    else {
		//then the best we can do is search the whole file
		return getIterator();
	    }
	}

	return it;
    }

    //get an iterator over the data file
    public RecordIterator getIterator()
	throws BufferManagerException, DiskManagerException, IOException
    {
	return mainfile.getIterator();
    }

    /* This class implements an iterator which allows the searching of
     * searchkeys, using an index on the schema's of the searchkey */
    private static class IndexIterator implements RecordIterator
    {
	protected MainFile file;
	protected Index index;
	protected Record keyvalue; //the value we are looking for inside the index
	protected RecordAddressIterator it;
	protected boolean bHasNext;
	protected RecordAddress currentAddr;
	protected Record currentRecord;
	
	public IndexIterator(MainFile file, Index index, Record searchkey)
	    throws BufferManagerException, DiskManagerException, IOException, SchemaException
	{
	    this.file = file;
	    this.index = index;
	    this.keyvalue = searchkey.project(index.getSchema());
	    this.it = index.findRecord(keyvalue);

	    bHasNext = false;
	    try {
		if (it.hasNext()) {
		    currentAddr = it.next();
		    currentRecord = file.getRecord(currentAddr);
		    bHasNext = true;
		}
		
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	}

	public boolean hasNext()
	{
	    return bHasNext;
	}

	public Record next()
	    throws NoSuchElementException, IOException, BufferManagerException, DiskManagerException
	{
	    if (!hasNext())
		throw new NoSuchElementException();

	    Record result = currentRecord;

	    bHasNext = false;
	    try {
		if (it.hasNext()) {
		    currentAddr = it.next();
		    currentRecord = file.getRecord(currentAddr);
		    bHasNext = true;
		}
		
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	    return result;
	}

	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    if (currentAddr == null)
		throw new NoSuchElementException();
	    
	    return currentAddr;
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) 
	    throws  SchemaException, BufferManagerException, DiskManagerException, IOException
	{
	    if (currentAddr == null)
		throw new NoSuchElementException();

	    
	    file.updateRecord(currentAddr, rec);
	}

	/* closes the attached resources. Calling other methods after
	 * a call to close() can fail */
	public void close()
	{
	    bHasNext = false;
	    it.close();
	    currentAddr = null;
	    currentRecord = null;
	}

	/* if the iterator is garbage-collected we should close to free the associated resources */
	public void finalize()
	{
	    close();
	}
	
    }

    /* This class implements an iterator which allows the searching of
     * searchkeys, using an index on subschema's of the searchkey */
    private static class SubIndexIterator implements RecordIterator
    {
	protected MainFile file;
	protected Index index;
	protected Record keyvalue; //the value we are looking for inside the index
	protected Record searchkey; //the key value we are really looking for
	protected RecordAddressIterator it;
	protected boolean bHasNext;
	protected RecordAddress currentAddr;
	protected Record currentRecord;
	
	public SubIndexIterator(MainFile file, Index index, Record searchkey)
	    throws BufferManagerException, DiskManagerException, IOException, SchemaException
	{
	    this.file = file;
	    this.index = index;
	    this.searchkey = searchkey;
	    this.keyvalue = searchkey.project(index.getSchema());
	    this.it = index.findRecord(keyvalue);

	    bHasNext = false;
	    try {
		while (!bHasNext && it.hasNext()) {
		    currentAddr = it.next();
		    currentRecord = file.getRecord(currentAddr);
		    if (currentRecord.project(searchkey.getSchema()).equals(searchkey))
			bHasNext = true;
		}
		
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	    catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	}

	public boolean hasNext()
	{
	    return bHasNext;
	}

	public Record next()
	    throws NoSuchElementException, IOException, BufferManagerException, DiskManagerException
	{
	    if (!hasNext())
		throw new NoSuchElementException();

	    Record result = currentRecord;

	    bHasNext = false;
	    try {
		while (!bHasNext && it.hasNext()) {
		    currentAddr = it.next();
		    currentRecord = file.getRecord(currentAddr);
		    if (currentRecord.project(searchkey.getSchema()).equals(searchkey))
			bHasNext = true;
		}
		
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	    catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	    return result;
	}

	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    if (currentAddr == null)
		throw new NoSuchElementException();

	    return currentAddr;
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) 
	    throws  SchemaException, BufferManagerException, DiskManagerException, IOException
	{
	    if (currentAddr == null)
		throw new NoSuchElementException();

	    file.updateRecord(currentAddr, rec);
	}

	/* closes the attached resources. Calling other methods after
	 * a call to close() can fail */
	public void close()
	{
	    bHasNext = false;
	    it.close();
	    currentAddr = null;
	    currentRecord = null;
	}

	/* if the iterator is garbage-collected we should close to free the associated resources */
	public void finalize()
	{
	    close();
	}
	
    }
}
