/*-------------------------------------------------------------------
   Authors: 2001-2003: Stijn Vansummeren

   The BufferManager class manages the memory buffers (which are
   ordinary sequences of bytes). The BufferManager can read/write a
   disk block into a buffer, and link the buffer to a Java Block
   object, which has methods to operate easier on the byte sequence.

   The BufferManager class provides capabilities to:
   - Allocate a free block 
   - Deallocate a block 
   - Pin a block : read a disk block from secondary storage into a buffer
     and link te buffer to a Java Block object
   - Unpin a block: return the buffer encapsulated by a Java block object
     to the buffer pool, allowing it to be reused.
     It's contents will be written to disk when the buffer is reused. 
   - Flush a buffer to disk
   - Flush all buffers to disk
 
   It uses a DiskManager to do all the I/O work.

   $Id: BufferManager.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db.buffermanager;


import java.lang.*;
import java.lang.ref.*;
import java.util.*;
import java.nio.*;
import java.io.*;

import be.ulb.db.*;
import be.ulb.db.diskmanager.*;

public class BufferManager
{
    /* The minimum number of buffer we always allocate */
    protected static int MINBUFFPOOL = 10;
    
    protected DiskManager diskmanager;
    protected int numbuffers;
    protected ByteBuffer [] buffpool;  //contains the true buffpool (sequences of bytes)
    protected Hashtable addr_to_buffnum; // maps BlockAddress objects to buffer numbers 
    protected BlockAddress [] buffnum_to_addr; //maps a buffer index to the address of the block it contains
    protected HashSet [] buffnum_to_blocks; //maps a buffer index to a linked list of blocks wrapping the buffer

    protected boolean [] dirty; //dirty[buffnum] == true when buffer at index buffnum should be writen back to disk
    protected boolean [] dofree; //dofree[buffnum] == true when buffer at index buffnum should be freed when its pincount reaches zero

    protected BufferReplacementPolicy policy;

    /* create a buffer manager with numbuffers buffers, using
     * DiskManger diskmanager, and the clock replacement policy */
    public BufferManager(int numbuffers, DiskManager diskmanager)
    {
	this(numbuffers, diskmanager, null);
	policy = new ClockBufferReplacementPolicy(this.numbuffers);
    }

    /* create a buffer manager with numbuffers buffers, using
     * DiskManger diskmanager, and the policy replacement policy */
    public BufferManager(int numbuffers, DiskManager diskmanager, BufferReplacementPolicy policy)
    {
	this.diskmanager = diskmanager;
	
	/* We always allocate NUMBUFFERS buffpool  */
	this.numbuffers = numbuffers < MINBUFFPOOL ? MINBUFFPOOL : numbuffers;
	buffpool = new ByteBuffer[this.numbuffers];
	/* set up the hashtable which maps BlockAddress objects to buffer numbers */
	addr_to_buffnum = new Hashtable(this.numbuffers);
	buffnum_to_addr = new BlockAddress[this.numbuffers];
	buffnum_to_blocks = new HashSet[this.numbuffers];
	dirty = new boolean[this.numbuffers];
	dofree = new boolean[this.numbuffers];

	for (int i = 0; i < this.numbuffers; i++) {
	    /* allocate a bytebuffer to hold the byte sequence of a block on disk in memory.
	       We allocate DIRECT buffpool for greater performance */
	    buffpool[i] = ByteBuffer.allocateDirect(diskmanager.getBlockSize());

	    /* allocate a HashSet for each buffer index in buffnum_to_blocks*/
	    buffnum_to_blocks[i] = new HashSet();
	}
	

	/* set up the replacement policy */
	this.policy = policy;

	/* Make sure that all buffer get written to disk when the Java VM stops */
	Runtime.getRuntime().addShutdownHook(new CleanupThread(this));
	
// 	System.out.println("Buffer manager created. Number of buffer = " + this.numbuffers +
// 			   ". Buffer size = " + diskmanager.getBlockSize());
    }

    /* Pin a block with address addr : read a disk block from
       secondary storage into a buffer and link te buffer to a Java
       Block object. The kind of the returned java block object is
       determined by the BlockFactory factory, which is repsonsible
       for actually creating the Java block. */
    public Block pinBlock(BlockAddress addr, BlockFactory factory)
	throws IOException, DiskManagerException, BufferManagerException
    {
// 	System.out.println("Pinning block at offset " + addr.getBlockOffset());
	
	/* first check if we don't already have it in memory */
	Integer buffnum = (Integer) addr_to_buffnum.get(addr);
	ByteBuffer buff;
	
	if (buffnum != null) 
	    //check if the requested block has been freed
	    if (dofree[buffnum.intValue()])
		throw new BufferManagerException("Block " + addr.getBlockOffset() + " is not allocated (has been freed).");
	    else {
		/* no need to read the buffer back in - it's already in memory */
		buff = buffpool[buffnum.intValue()];
		
// 		System.out.println("Buffer was already in memory!");
	    }
	else {
	    /* It's not in memory. Find a free buffer */
	    try {
		buffnum = new Integer(policy.pickVictim());
	    } 
	    catch (BufferReplacementPolicyNoVictimException e) {
		throw new BufferManagerException("Cannot pin block - buffer pool exhausted.");
	    }

// 	    System.out.println("Buffer was not in memory - replacement victim = " + buffnum.intValue());
	    
	    buff = buffpool[buffnum.intValue()];

	    /* remove the associated block Address from the
	     * addr_to_buffnum mapping and write the contents back to
	     * disk if the block was modified */
	    BlockAddress oldaddr = buffnum_to_addr[buffnum.intValue()];
	    if (oldaddr != null) {
		addr_to_buffnum.remove(oldaddr);
		if (dirty[buffnum.intValue()]) {
// 		    System.out.println("Replacement buffer was dirty - writing it back to disk");
		    buff.limit(buff.capacity());
		    buff.position(0);
		    diskmanager.writeBlock(buff, oldaddr);
		}
	    }
	    
	    buffnum_to_addr[buffnum.intValue()] = addr;
	    addr_to_buffnum.put(addr, buffnum);

	    //this buffer is certainly not marked to be freed and cannot be dirty
	    dofree[buffnum.intValue()] = false;
	    dirty[buffnum.intValue()] = false;
	    
	    /* finally read the contents from disk */
	    buff.limit(buff.capacity());
	    buff.position(0);
	    diskmanager.readBlock(buff, addr);
	}
	
	policy.pin(buffnum.intValue());
// 	System.out.println("Buffer " + buffnum + " pinned. Pincount = " + policy.getPinCount(buffnum.intValue()));

	buff.limit(buff.capacity());
	buff.position(0);
	Block result = factory.create(buff, addr);
	
	/* maintain a list of the blocks wrapped around this buffer */
	buffnum_to_blocks[buffnum.intValue()].add(result);

	return result;
    }
    
    /* Unpin a block: return the buffer encapsulated by a Java block
     * object to the buffer pool, allowing it to be reused. If the
     * block is marked to be freed we actually free the block when its
     * pincount reaches zero.

     * If block.isModified() == true then the buffers contents will be
     * written to disk when the buffer is reused. */
    public void unpinBlock(Block block) throws BufferManagerException, DiskManagerException, IOException
    {
	/* look up the buffer number */
	BlockAddress addr = block.getBlockAddress();
	Integer buffnum = (Integer) addr_to_buffnum.get(addr);
	
	if (buffnum == null)
	    throw new BufferManagerException("Block was not pinned!");

	/* detach the block from the buffer - this way exceptions get
	 * thrown if the block is subsequently used */
	block.detach();

	/* also remove it from our list of blocks wrapping the buffer */
	if (!buffnum_to_blocks[buffnum.intValue()].remove(block))
	    throw new BufferManagerException("Bock was not pinned!");

	/* the acutal unpinning ... */
	ByteBuffer buff = buffpool[buffnum.intValue()];
	dirty[buffnum.intValue()] |= block.isModified();

	/* unpin the buffer */
	try {
	    policy.unpin(buffnum.intValue());

	    /* if there is no other in-memory block pinning this
	     * buffer and the buffer is marked to be freed,
	     * actually free it */
	    if (dofree[buffnum.intValue()] && (policy.getPinCount(buffnum.intValue()) == 0))
		truelyFree(buffnum.intValue());
		
// 	    System.out.println("Buffer number " + buffnum + "(" + addr.getBlockOffset() + ") succesfully unpinned.");
	} 
	catch (BufferReplacementPolicyException e) {
	    throw new BufferManagerException("Bock was not pinned!");
	}

    }

    /* allocates a new block on disk, and pins a buffer for it. The
     * contents of the buffer can be arbitrary data! */
    public Block newBlock(BlockFactory factory) throws BufferManagerException, DiskManagerException, IOException
    {
	BlockAddress addr = diskmanager.allocateBlock();
// 	System.out.println("Address of newly allocated block = " + addr.getBlockOffset());
	
	Block block = pinBlock(addr, factory);
	/* give the block a chance to initialize itself with some default contents.*/
	block.initContents();
	return block;
    }

    /* Frees a block on disk so that it's space can be reused - its
     * buffer is also reclaimed. Since there can be multiple Java
     * block object pinning the same buffer, we only truely free the
     * block when its pincount reaches zero. */
    public void freeBlock(Block block) throws BufferManagerException, DiskManagerException, IOException
    {
	BlockAddress addr = block.getBlockAddress();
	Integer buffnum = (Integer) addr_to_buffnum.get(addr);
	if (buffnum == null)
	    throw new BufferManagerException("Block to be freed is not in memory!");

	dofree[buffnum.intValue()] = true;
	unpinBlock(block);
    }

    private void truelyFree(int buffnum)
	throws BufferReplacementPolicyException, DiskManagerException, IOException
    {
	policy.free(buffnum);		    

	/* make sure we never write the buffer out again */
	BlockAddress addr = buffnum_to_addr[buffnum];
	buffnum_to_addr[buffnum] = null; 
	addr_to_buffnum.remove(addr);
	dofree[buffnum] = false;

	/* free it on disk */
	diskmanager.freeBlock(addr);

// 	System.out.println("Block at offset " + addr.getBlockOffset() + " truely freed.");
    }
    

    /* Force the contents of a block to disk */
    public void flushBlock(Block block) throws BufferManagerException, DiskManagerException, IOException
    {
	BlockAddress addr = block.getBlockAddress();

	Integer buffnum = (Integer) addr_to_buffnum.get(addr);
	if (buffnum == null)
	    throw new BufferManagerException("Block to be flushed is not validly attached to a buffer");

	ByteBuffer buff = buffpool[buffnum.intValue()];

	buff.limit(buff.capacity());
	buff.position(0);

	diskmanager.writeBlock(buff, addr);
    }

    
    /* flushes all in-memory buffpool to disk. If onlydirty == true we
     * only flush the dirty blocks - otherwise we flush them all */
    public void flushAll(boolean onlydirty) throws DiskManagerException, IOException
    {
	for (int i = 0; i < numbuffers; i++) 
	    /* do not write unless we have read it at least once */
	    if (buffnum_to_addr[i] != null) {
		if (!onlydirty || (dirty[i])) {
		    ByteBuffer buff = buffpool[i];
		    BlockAddress addr = buffnum_to_addr[i];
		    buff.position(0);
		    buff.limit(buff.capacity());
		    diskmanager.writeBlock(buff, addr);
		}
	    }
    }


    /* if the buffer manager is garbage-collected, we need to do some clean-up */
    public void finalize()
    {
	try {
	    //only flush the dirty buffers
	    flushAll(true);

	    for (int i = 0; i < numbuffers; i++) {
		//make sure blocks which are pending to be freed are
		//actually freed on disk
		if (dofree[i] && buffnum_to_addr[i] != null)
		    diskmanager.freeBlock(buffnum_to_addr[i]);

		Iterator it = buffnum_to_blocks[i].iterator();
		while (it.hasNext()) {
		    Block block = (Block) it.next();

		    //make sure the block will not be used again
		    block.detach();
		}
	    }
	}
	catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}
    }

    /* The following class is used to make sure all blocks are
     * correctly written to disk when the Java VM exits */
    private static class CleanupThread extends Thread
    {
	/* we use a weak reference so that a BufferManager can be
	   finalized and reclaimed before the Java VM stops */
	private WeakReference ref;
	
	public CleanupThread(BufferManager manager)
	{
	    ref = new WeakReference(manager);
	}

	public void run() 
	{
	    BufferManager manager = (BufferManager) ref.get();
	    if (manager != null)
		try {
		    manager.finalize();
		}
		catch(Exception e) {
		    e.printStackTrace();
		}
	}
    }
}
