/*-------------------------------------------------------------------
   Authors: 2003: Stijn Vansummeren

   The BufferReplacementPolicy class represents a policy to decide
   which unpinned in-memory buffer should be written back to disk when
   we need to replace a buffer (for instance because we need to read
   in a block which is currently not in memory).  Many policies exist
   (least recently used, clock, ...). This is a base class for
   concrete policies.

   $Id: BufferReplacementPolicy.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db.buffermanager;

public abstract class BufferReplacementPolicy
{
    /* keeps the pincount of each buffer */
    protected int [] pincounts;
    protected int numbuffers;
    
    public BufferReplacementPolicy(int numbuffers)
    {
	pincounts = new int[numbuffers];
	this.numbuffers = numbuffers;
    }
    
    /* increase the pin count of the buffer with index buffnum. Pinned
     * buffer will never be chosen as a candidate for replacement */
    public void pin(int buffnum)
    {
	(pincounts[buffnum])++;
    }
    
    /* unpin the buffer: indicate that the user does not need the
     * buffer any more. Buffers with pincount = 0 are possible
     * candidates for replacements */
    public void unpin(int buffnum) throws BufferReplacementPolicyException
    {
	if (pincounts[buffnum] > 0)
	    (pincounts[buffnum])--;
	else
	    throw new BufferReplacementPolicyException("Buffer " + buffnum + "was not pinned");
    }

    /* free a buffer - should only be used when the buffer's content
       is also deallocated on disk.  Freed buffers are always eligble
       for replacement. */
    public void free(int buffnum) throws BufferReplacementPolicyException
    {
	if (pincounts[buffnum] > 1)
	    throw new BufferReplacementPolicyException("Cannot free buffer " + buffnum +
						 " - it's pincount = " + pincounts[buffnum] + " > 1");
	
	pincounts[buffnum] = 0;
    }

    /* returns the buffer number which should be written back to disk
     * (if it's dirty) and whose content can then be reused. The
     * victims returned are never pinned. */
    public abstract int pickVictim() throws BufferReplacementPolicyNoVictimException ;
    

    /* returns the pincount of a buffer */
    public int getPinCount(int buffnum)
    {
	return pincounts[buffnum];
    }

    
}


