/*-------------------------------------------------------------------
  Authors: 2003: Stijn Vansummeren

  The attribute catalog records information about the attributes of
  the various relations found in a database.

  $Id: AttributeCatalog.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db.catalog;


import java.io.*;
import java.lang.*;
import java.util.*;
import java.lang.ref.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.file.*;
import be.ulb.db.schema.*;

public class AttributeCatalog
{
    protected BufferManager bm;
    protected MainFile file;

    private static final Schema schema = new AttributeCatalogSchema();

    /* we maintain weak references to our spawned iterators so that we
     * can close them when the file is closed */
    private List iterators;

    /* creates a new relation catalog */
    public AttributeCatalog(BufferManager bm)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this(bm, null);
    }
    
    /* loads an existing relation catalog */
    public AttributeCatalog(BufferManager bm, BlockAddress root)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.bm = bm;
	this.iterators = new LinkedList();
	
	if (root == null)
	    file = new MainFile(bm, schema);
	else
	    file = new MainFile(bm, schema, root);
    }

    /* returns the address of the block from which the catalog can be loaded */
    public BlockAddress getBlockAddress()
    {
	return file.getBlockAddress();
    }
    

    /* add schema information for relation name to the catalog. Throws
     * IllegalArgumentException if there already is schema information
     * for this relation. */
    public void add(String name, Schema attrschema)
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordIterator it = find(name);
	if (it.hasNext())
	    throw new IllegalArgumentException("There already exists a schema for relation " + name);
	it.close();

	Record rec = new Record(schema);
	try {
	    rec.putString(AttributeCatalogSchema.relNameIndex, name);
	    for (int i = 0; i < attrschema.length(); i++) {
		rec.putString(AttributeCatalogSchema.attrNameIndex, attrschema.getName(i));
		rec.putString(AttributeCatalogSchema.typeIndex, attrschema.getType(i).getClass().getName());
		rec.putInt(AttributeCatalogSchema.lengthIndex, new Integer(attrschema.getType(i).getLength()));
		rec.putInt(AttributeCatalogSchema.offsetIndex, new Integer(i));

		file.addRecord(rec);
	    }
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
    }

    /* remove all schema information about relation name */
    public void remove(String name)
	throws BufferManagerException, DiskManagerException, IOException, NoSuchElementException
    {
	RecordIterator it = getIterator();
	
	try {
	    while (it.hasNext()) {
		Record rec = it.next();
		if (rec.getString(AttributeCatalogSchema.relNameIndex).equals(name))
		    file.removeRecord(it.getRecordAddress());
	    }
	}
	catch(IndexOutOfBoundsException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
		
	it.close();
    }
    

    /* returns a record iterator to all records having relation name
     * name. */
    public RecordIterator find(String name)
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordIterator result = new AttrIterator(file, name);
	WeakReference ref = new WeakReference(result);
	iterators.add(ref);
	return result;
    }

    /* returns the record havin having relation name
     * relname and attribute name attrname */
    public Record find(String relname, String attrname)
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordIterator it = find(relname);
	boolean found = false;

	try {
	    Record rec = null;
	    while (!found && it.hasNext()) {
		rec = it.next();
		String aname = rec.getString(AttributeCatalogSchema.attrNameIndex);
		found = attrname.equals(aname);
	    }

	    if (found)
		return rec;
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}

	it.close();
    return null;
    }


    private static class AttrIterator implements RecordIterator
    {
	private MainFile file;
	private RecordIterator it;
	private String name;
	private boolean bHasNext;
	private RecordAddress current;
	
	public AttrIterator(MainFile file, String name)
	    throws BufferManagerException, DiskManagerException, IOException
	{
	    this.file = file;
	    this.it = file.getIterator();
	    this.name = name;

	    bHasNext = false;
	    try {
		while (!bHasNext && it.hasNext())
		    bHasNext = it.next().getString(AttributeCatalogSchema.relNameIndex).equals(name);
		
		if (bHasNext)
		    current = it.getRecordAddress();
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	    catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	}

	public boolean hasNext()
	{
	    return bHasNext;
	}

	public Record next()
	    throws NoSuchElementException, IOException, BufferManagerException, DiskManagerException
	{
	    if (!hasNext())
		throw new NoSuchElementException();

	    Record result = file.getRecord(current);

	    bHasNext = false;
	    try {
		while (!bHasNext && it.hasNext())
		    bHasNext = it.next().getString(AttributeCatalogSchema.relNameIndex).equals(name);
		
		if (bHasNext)
		    current = it.getRecordAddress();
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	    catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	    return result;
	}

	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    return it.getRecordAddress();
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) 
	    throws  SchemaException, BufferManagerException, DiskManagerException, IOException
	{
	    it.updateRecord(rec);
	}

	/* closes the attached resources. Calling other methods after
	 * a call to close() can fail */
	public void close()
	{
	    bHasNext = false;
	    it.close();
	}

	/* if the iterator is garbage-collected we should close to free the associated resources */
	public void finalize()
	{
	    close();
	}
	
    }

    /* returns an iterator to all records in the catalog */
    public RecordIterator getIterator()
	throws BufferManagerException, DiskManagerException, IOException
    {
	return file.getIterator();
    }

    //closes the catalog an frees the associated resources. Subsequent operations may fails
    public void close()
    {
	//close all spawned iterators
	ListIterator it = iterators.listIterator();
	while (it.hasNext()) {
	    WeakReference ref = (WeakReference) it.next();
	    RecordIterator rit = (RecordIterator) ref.get();
	    if (rit != null)
		rit.close();
	}
	iterators.clear();

	file.close();
    }

    //close when we get garbage-collected
    public void finalize()
    {
	close();
    }
}


