/*-------------------------------------------------------------------
  Authors: 2001-2003: Stijn Vansummeren

  The database catalog describes the contents of a database:
  - which relations there are
  - the schema of each relation
  - the indexes on relations
  - information needed for query optimization (how many tuples, ...)

  $Id: Catalog.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db.catalog;



import java.io.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.file.*;
import be.ulb.db.schema.*;

public class Catalog
{
    protected BufferManager bm;
    protected BlockAddress rootAddr;

    protected RelationCatalog relCatalog;
    protected AttributeCatalog attrCatalog;
    protected IndexCatalog indexCatalog;
    protected IndexAttrCatalog indexAttrCatalog;

    /* describes the records in the catalog root block, which are
     * simple pointers to the other catalogs */
    private static class InfoSchema extends Schema
    {
	private static String ADDR = "addr";
	
	public InfoSchema()
	{
	    try {
		String [] names =  {ADDR};
		AttributeType [] types = {new AttributeTypeBlockAddress()};
		
		initSchema(names, types);
	    } catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	}

	/* statically determine the offset of the attributes */
	public static int addrIndex = 0;
    }
    
    protected static final Schema infoSchema = new InfoSchema();
    protected static final int RELSLOT = 0; //position of the pointer to the relation catalog in the root catalog block
    protected static final int ATTRSLOT = 1; //position of the pointer to the attribute catalog in the root catalog block
    protected static final int INDEXSLOT = 2; //position of the pointer to the index catalog in the root catalog block
    protected static final int INDEXATTRSLOT = 3; //position of the pointer to the indexattr catalog in the root catalog block

    /** Loads the catalog */
    public Catalog(BufferManager bm, DiskManager dm) 
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.bm = bm;

	//the first block of the catalog is stored in the DB Meta
	//Block of the disk manager
	rootAddr = dm.getDBMetaBlockAddress();
	MainBlock block = (MainBlock) bm.pinBlock(rootAddr, new MainBlockFactory(infoSchema));

	try {
	    //do we need to create a new catalog?
	    if (block.size() == 0) {
		relCatalog = new RelationCatalog(bm);
		attrCatalog = new AttributeCatalog(bm);
		indexCatalog = new IndexCatalog(bm);
		indexAttrCatalog = new IndexAttrCatalog(bm);

		Record rec = new Record(infoSchema);
	    
		rec.put(InfoSchema.addrIndex, relCatalog.getBlockAddress());
		block.addRecord(rec);
	    
		rec.put(InfoSchema.addrIndex, attrCatalog.getBlockAddress());
		block.addRecord(rec);
	    
 		rec.put(InfoSchema.addrIndex, indexCatalog.getBlockAddress());
 		block.addRecord(rec);

		rec.put(InfoSchema.addrIndex, indexAttrCatalog.getBlockAddress());
 		block.addRecord(rec);

	    } else {
		Record rec = block.getRecord(RELSLOT);
		relCatalog = new RelationCatalog(bm, rec.getBlockAddress(InfoSchema.addrIndex));
	    
		rec = block.getRecord(ATTRSLOT);
		attrCatalog = new AttributeCatalog(bm, rec.getBlockAddress(InfoSchema.addrIndex));
	    
 		rec = block.getRecord(INDEXSLOT);
 		indexCatalog = new IndexCatalog(bm, rec.getBlockAddress(InfoSchema.addrIndex));

 		rec = block.getRecord(INDEXATTRSLOT);
 		indexAttrCatalog = new IndexAttrCatalog(bm, rec.getBlockAddress(InfoSchema.addrIndex));
	    }
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); //should not happen
	}
	catch(BlockFullException e) {
	    e.printStackTrace();
	    System.exit(1); //should not happen
	}
	
	bm.unpinBlock(block);
    }

    /* closes the catalog */
    public void close()
    {
	if (relCatalog != null)
	    relCatalog.close();
	
	if (attrCatalog != null)
	    attrCatalog.close();
	
	if (indexCatalog != null)
	    indexCatalog.close();
    }

    /* always close when we get garbage collected */
    public void finalize()
    {
	close();
    }
    
    public RelationCatalog getRelationCatalog()
    {
	return relCatalog;
    }
    
    public AttributeCatalog getAttributeCatalog()
    {
	return attrCatalog;
    }

    public IndexCatalog getIndexCatalog()
    {
	return indexCatalog;
    }

    public IndexAttrCatalog getIndexAttrCatalog()
    {
	return indexAttrCatalog;
    }
        

}
