/*-------------------------------------------------------------------
  Authors: 2003: Stijn Vansummeren

  The index catalog records information about the index on the various
  relations found in a database.

  $Id: IndexCatalog.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db.catalog;


import java.io.*;
import java.lang.*;
import java.util.*;
import java.lang.ref.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.file.*;
import be.ulb.db.index.*;
import be.ulb.db.schema.*;

public class IndexCatalog
{
    protected BufferManager bm;
    protected MainFile file;

    private static final Schema schema = new IndexCatalogSchema();

    /* we maintain weak references to our spawned iterators so that we
     * can close them when the indexcatalog is closed */
    private List iterators;

    /* creates a new relation catalog */
    public IndexCatalog(BufferManager bm)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this(bm, null);
    }
    
    /* loads an existing relation catalog */
    public IndexCatalog(BufferManager bm, BlockAddress root)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.bm = bm;
	this.iterators = new LinkedList();

	if (root == null) {
	    file = new MainFile(bm, schema);
	}
	else {
	    file = new MainFile(bm, schema, root);
	}
	
    }

    /* returns the address of the block from which the catalog can be loaded */
    public BlockAddress getBlockAddress()
    {
	return file.getBlockAddress();
    }


    /* add an index index relation relname, with identifier
     * idxid. Throws IllegalArgumentException if there already is an
     * index with the given id for this relation. */
    public void add(String relname, Index index, int idxid)
	throws BufferManagerException, DiskManagerException, IOException
    {
	try {
	    RecordIterator it = find(relname);
	    while (it.hasNext()) {
		Record rec = it.next();
		String name = rec.getString(IndexCatalogSchema.relNameIndex);
		int id = rec.getInt(IndexCatalogSchema.idIndex).intValue();
		if (id == idxid && name.equals(relname))
		    throw new IllegalArgumentException("There already is an index " + id + " for relation " + relname);
	    }
	    it.close();

	    Record rec = new Record(schema);
	    rec.putString(IndexCatalogSchema.relNameIndex, relname);
	    rec.putString(IndexCatalogSchema.typeIndex, index.getClass().getName());
	    rec.putInt(IndexCatalogSchema.idIndex, new Integer(idxid));
	    rec.putInt(IndexCatalogSchema.attrNumIndex, new Integer(index.getSchema().length()));
	    rec.putBlockAddress(IndexCatalogSchema.addrIndex, index.getBlockAddress());
	    file.addRecord(rec);
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
    }

    /* remove all index information of relation relname */
    public void remove(String relname)
	throws BufferManagerException, DiskManagerException, IOException, NoSuchElementException
    {
	RecordIterator it = getIterator();
	
	try {
	    while (it.hasNext()) {
		Record rec = it.next();
		String name = rec.getString(IndexCatalogSchema.relNameIndex);
		if (name.equals(relname))
		    file.removeRecord(it.getRecordAddress());
	    }
	}
	catch(IndexOutOfBoundsException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
		
	it.close();
    }


    /* remove index information with id idxid on relation relname */
    public void remove(String relname, int idxid)
	throws BufferManagerException, DiskManagerException, IOException, NoSuchElementException
    {
	RecordIterator it = getIterator();
	
	try {
	    while (it.hasNext()) {
		Record rec = it.next();
		String name = rec.getString(IndexCatalogSchema.relNameIndex);
		int id = rec.getInt(IndexCatalogSchema.idIndex).intValue();
		if (id == idxid && name.equals(relname))
		    file.removeRecord(it.getRecordAddress());
	    }
	}
	catch(IndexOutOfBoundsException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
		
	it.close();
    }
    

    /* returns a record iterator to all records having relation name
     * name. */
    public RecordIterator find(String name)
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordIterator result = new IdxIterator(file, name);
	WeakReference ref = new WeakReference(result);
	iterators.add(ref);
	return result;
    }

    private static class IdxIterator implements RecordIterator
    {
	private MainFile file;
	private RecordIterator it;
	private String name;
	private boolean bHasNext;
	private RecordAddress current;
	
	public IdxIterator(MainFile file, String name)
	    throws BufferManagerException, DiskManagerException, IOException
	{
	    this.file = file;
	    this.it = file.getIterator();
	    this.name = name;

	    bHasNext = false;
	    try {
		while (!bHasNext && it.hasNext())
		    bHasNext = it.next().getString(IndexCatalogSchema.relNameIndex).equals(name);
		
		if (bHasNext)
		    current = it.getRecordAddress();
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	    catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	}

	public boolean hasNext()
	{
	    return bHasNext;
	}

	public Record next()
	    throws NoSuchElementException, IOException, BufferManagerException, DiskManagerException
	{
	    if (!hasNext())
		throw new NoSuchElementException();

	    Record result = file.getRecord(current);

	    bHasNext = false;
	    try {
		while (!bHasNext && it.hasNext())
		    bHasNext = it.next().getString(IndexCatalogSchema.relNameIndex).equals(name);
		
		if (bHasNext)
		    current = it.getRecordAddress();
	    }
	    catch(NoSuchElementException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	    catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	    return result;
	}

	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    return it.getRecordAddress();
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) 
	    throws  SchemaException, BufferManagerException, DiskManagerException, IOException
	{
	    it.updateRecord(rec);
	}

	/* closes the attached resources. Calling other methods after
	 * a call to close() can fail */
	public void close()
	{
	    bHasNext = false;
	    it.close();
	}

	/* if the iterator is garbage-collected we should close to free the associated resources */
	public void finalize()
	{
	    close();
	}
	
    }

    /* returns an iterator to all records in the catalog */
    public RecordIterator getIterator()
	throws IOException, BufferManagerException, DiskManagerException
    {
	return file.getIterator();
    }

    //closes the catalog an frees the associated resources. Subsequent operations may fails
    public void close()
    {
	//close all spawned iterators
	ListIterator it = iterators.listIterator();
	while (it.hasNext()) {
	    WeakReference ref = (WeakReference) it.next();
	    RecordIterator rit = (RecordIterator) ref.get();
	    if (rit != null)
		rit.close();
	}
	iterators.clear();

	file.close();
    }

    //close when we get garbage-collected
    public void finalize()
    {
	close();
    }

}


