/*-------------------------------------------------------------------
  Authors: 2003: Stijn Vansummeren

  The relation catalog records information about the relations found
  in a database.

  $Id: RelationCatalog.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db.catalog;


import java.io.*;
import java.lang.*;
import java.util.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.file.*;
import be.ulb.db.schema.*;

public class RelationCatalog
{
    protected BufferManager bm;
    protected  MainFile file;

    private static final Schema schema = new RelationCatalogSchema();

    /* creates a new relation catalog */
    public RelationCatalog(BufferManager bm)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this(bm, null);
    }
    
    /* loads an existing relation catalog */
    public RelationCatalog(BufferManager bm, BlockAddress root)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.bm = bm;
	if (root == null)
	    file = new MainFile(bm, schema);
	else
	    file = new MainFile(bm, schema, root);
    }

    /* returns the address of the block from which the catalog can be loaded */
    public BlockAddress getBlockAddress()
    {
	return file.getBlockAddress();
    }

    /* add a new relation to the catalog. Throws
     * IllegalArgumentException if there already is a relation with
     * the same name */
    public void add(String name, BlockAddress addr, Schema relschema)
	throws BufferManagerException, DiskManagerException, IOException
    {
	if (find(name) != null)
	    throw new IllegalArgumentException("There already exists a relation " + name);

	Record rec = new Record(schema);
	try {
	    rec.putString(RelationCatalogSchema.nameIndex, name);
	    rec.putBlockAddress(RelationCatalogSchema.addrIndex, addr);
	    rec.putInt(RelationCatalogSchema.attrNumIndex, new Integer(relschema.length()));
	    file.addRecord(rec);
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
    }

    public void remove(String name)
	throws BufferManagerException, DiskManagerException, IOException, NoSuchElementException
    {
	RecordAddress pos = internal_find(name);
	if (pos == null)
	    throw new NoSuchElementException();
	
	file.removeRecord(pos);
    }

    /* returns a record adhering to RelationCatalogSChema if there is
     * a relation with name name, and null otherwise */
    public Record find(String name)
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordAddress pos = internal_find(name);
	if (pos == null)
	    return null;

	Record result = null;
	try {
	    result =  file.getRecord(pos);
	}
	catch(IndexOutOfBoundsException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	
	return result;
    }

//         /* updates the address where the relation with name name
//      * starts. Throws NoSuchElementException if there is no relation
//      * with that name. */
//     public void update(String name, BlockAddress addr)
// 	throws BufferManagerException, DiskManagerException, IOException, NoSuchElementException
//     {
// 	RecordAddress pos = internal_find(name);
// 	if (pos == null)
// 	    throw new NoSuchElementException();
	
// 	try {
// 	    Record rec = new Record(schema);
// 	    rec.put(RelationCatalogSchema.nameIndex, name);
// 	    rec.put(RelationCatalogSchema.addrIndex, addr);
	    
// 	    file.updateRecord(pos, rec);
// 	}
// 	catch (SchemaException e) {
// 	    e.printStackTrace();
// 	    System.exit(1); // it should not occur
// 	}
// 	catch (IndexOutOfBoundsException e) {
// 	    e.printStackTrace();
// 	    System.exit(1); // it should not occur
// 	}
//     }

    public RecordIterator getIterator()
	throws BufferManagerException, DiskManagerException, IOException
    {
	return file.getIterator();
    }

    //closes the catalog an frees the associated resources. Subsequent operations may fail
    public void close()
    {
	file.close();
    }

    //close when we get garbage-collected
    public void finalize()
    {
	close();
    }
    
    private RecordAddress internal_find(String name)
	throws IOException, BufferManagerException, DiskManagerException
    {
	RecordIterator it = file.getIterator();
	boolean found = false;

	RecordAddress result = null;
	try {
	    while (!found && it.hasNext()) {
		Record rec = it.next();
		found = rec.get(RelationCatalogSchema.nameIndex).equals(name);
	    }

	    if (found) 
		result = it.getRecordAddress();
	}
	catch (SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	catch (NoSuchElementException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	
	it.close();

	return result;
    }
}


