/*-------------------------------------------------------------------
   Authors: 2003: Stijn Vansummeren

   The DiskManager class provides capabilities to:
   - Create a new database on secondary storage
   - Destroy a database on secondary storage
   - Allocate a block on secondary storage
   - Free a block on secondary storage allowing it to be reused
   - Write a buffer to a block on secondary storage
   - Read a block from secondary storage into a buffer

   $Id: DiskManager.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db.diskmanager;

import java.lang.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;

import be.ulb.db.*;
import be.ulb.db.schema.*;


public class DiskManager
{
    protected static int STDBLOCKSIZE = 4096;
    protected static int MINBLOCKSIZE = 512;
    protected static int MINNUMBLOCKS = 10;
    
    /* The index (in blocks) of the block at which the disk manager
     * can store some meta-information (e.g. about free blocks) */
    protected static final int DMMETAINDEX = 0; 
    
    /* The index (in blocks) of the block at which the DBMS stores
       meta-information about the database (schema definitions,
       relation names and their start block, indexes and their start
       block, ...) */
    protected static final int DBMETAINDEX = 1;

    protected int numblocks = 0;
    protected int blocksize = 0;
    protected FileChannel channel;
    /* points to the first block of a linked list of free blocks */
    protected BlockAddress freeListHeadAddr; 
    /* Buffer for internal use */
    protected ByteBuffer buff;
    
    /* creates a new database with numblocks blocks of default block size. It throws
     * DiskManagerException if the database already exists. */
    public DiskManager(File file, int numblocks) throws IOException, DiskManagerException
    {
	this(file,numblocks, STDBLOCKSIZE);
    }

    /* creates a new database with numblocks blocks. It throws
     * DiskManagerException if the database already exists. */
    public DiskManager(File file, int numblocks, int  blocksize) throws IOException, DiskManagerException
    {
	if (file.exists()) 
	    throw new DiskManagerException("Cannot create a database that already exists");
	
	RandomAccessFile rafile = new RandomAccessFile(file, "rwd");
	channel = rafile.getChannel();

	/* The block size should be at least one kb */
	this.blocksize = blocksize < MINBLOCKSIZE ? MINBLOCKSIZE : blocksize;

	/* The database should contain at least MINUMBLOCKS blocks (needed to
	 * store meta-information) */
	this.numblocks = numblocks < MINNUMBLOCKS ? MINNUMBLOCKS : numblocks;

	/* create a buffer the size of a block */
	this.buff = ByteBuffer.allocateDirect(this.blocksize);

	/* Create an empty file. The first two blocks are used to
	 * store meta-information.*/

	/* We create a list of free blocks containing all blocks in
	 * the database except the ones used to store meta
	 * information*/
	for (int i = 2; i < this.numblocks; i++) {
	    BlockAddress addr = new BlockAddress(i * this.blocksize);
	    DiskManagerFreeListBlock block = new DiskManagerFreeListBlock(buff.duplicate(), addr);

	    /* We write an invalid "next" address in the last block */ 
	    BlockAddress next;
	    
	    if (i != this.numblocks - 1) 
		next = new BlockAddress((i+1) * this.blocksize);
	    else
		next = BlockAddress.INVALIDADDR;
	    
	    block.setNextBlockAddr(next);
	    buff.rewind(); /* rewind or not all data may be written */
	    writeBlock(buff, addr);
	}
	
	/* set the address of the first block of the linked list of free blocks */
	freeListHeadAddr = new BlockAddress(2 * this.blocksize);

	/* write the meta data to disk */
	writeMetaData();
    }

    /* Loads an existing database. It throws FileNotFoundException if
     * the database does not exists. */
    public DiskManager(File file) throws FileNotFoundException, IOException
    {
	if (!file.exists())
	    throw new FileNotFoundException();
	
	RandomAccessFile rafile = new RandomAccessFile(file, "rwd");
	channel = rafile.getChannel();

	/* read the meta data */
	readMetaData();
	
	this.buff = ByteBuffer.allocateDirect(blocksize);
	buff.rewind();
    }

    /* The following functions are responsible for writing and reading
       the metadata. The metadata starts at the beginning of file and
       is organised as follows: 
       - the first integer is the block size 
       - the second integer is the number of blocks
       - the third integer is the offset of the head of the free block list
    */

    protected static int HEADERSIZE = 12;

    protected void writeMetaData() throws IOException
    {
	ByteBuffer header = ByteBuffer.allocate(HEADERSIZE);
	header.putInt(blocksize);
	header.putInt(numblocks);
	header.putInt(freeListHeadAddr.getBlockOffset());

	header.rewind();
	channel.position(0);
	channel.write(header);
	channel.force(true);
    }

    protected void readMetaData() throws IOException
    {
	ByteBuffer header = ByteBuffer.allocate(HEADERSIZE);
	channel.position(0);
	channel.read(header);
	header.rewind();
	
	blocksize = header.getInt();
	numblocks= header.getInt();
	freeListHeadAddr = new BlockAddress(header.getInt());

	System.out.println("Meta data has been read");
	System.out.println("blocksize = " + blocksize);
	System.out.println("number of blocks = " + numblocks);
	System.out.println("free list starts at offset = " + freeListHeadAddr.getBlockOffset());
    }
    

    /* Returns the total number of blocks */
    public int getNumBlocks()
    {
	return numblocks;
    }

    /* Returns the number of bytes in a block */
    public int getBlockSize()
    {
	return blocksize;
    }

    /* Returns the block used to store meta information about the DB
     * (relation names, schema's, ...) */
    public BlockAddress getDBMetaBlockAddress()
    {
	return new BlockAddress(DBMETAINDEX * blocksize);
    }

        /* Write the buffer to the block on secondary storage with address
     * addr. */
    public void writeBlock(ByteBuffer buff, BlockAddress addr) throws DiskManagerException, IOException
    {
	/* Check if the bytebuffer has enough bytes to fill the block */
	if (buff.remaining() != blocksize)
	    throw new DiskManagerException("The number of bytes to be written differs from the block size.");

	/* go to the correct position in the file */
	channel.position(addr.getBlockOffset());
	channel.write(buff);
    }
    
    /* Read the block on secondary storage with address addr into the
     * buffer buff */
    public void readBlock(ByteBuffer buff, BlockAddress addr) throws DiskManagerException, IOException
    {
	/* Check if the bytebuffer has enough bytes to read in the entire block */
	if (buff.remaining() != blocksize)
	    throw new DiskManagerException("The number of bytes to be read differs from the block size. "
					   + blocksize + "<>" + buff.remaining());

	/* go to the correct position in the file */
	channel.position(addr.getBlockOffset());
	channel.read(buff);
    }

    /* Allocate a block on secondary storage. It returns the address
     * of the newly allocated block. */
    public BlockAddress allocateBlock() throws DiskManagerException, IOException
    {
	/* Make sure the head of our list is not the "end of list" address */
	if (!freeListHeadAddr.isValid())
	    throw new DiskManagerException("DB is full - cannot allocate new block!");

	BlockAddress oldHead = freeListHeadAddr;
	
	buff.rewind();
	readBlock(buff, oldHead);
	DiskManagerFreeListBlock block = new DiskManagerFreeListBlock(buff.duplicate(), oldHead);

	freeListHeadAddr = block.getNextBlockAddr();
	writeMetaData();
	
	return oldHead;
    }

    /* Free a block on secondary storage with address addr so that it
     * can be reused. */
    public void freeBlock(BlockAddress addr) throws DiskManagerException, IOException
    {
	buff.rewind();
	DiskManagerFreeListBlock block = new DiskManagerFreeListBlock(buff.duplicate(), addr);
	block.setNextBlockAddr(freeListHeadAddr);
	buff.rewind();
	writeBlock(buff,addr);
	freeListHeadAddr = addr;

	writeMetaData();
    }


}
