/*-------------------------------------------------------------------
   Authors: 2001 - 2003: Stijn Vansummeren

   The DirectoryFile is a linked list of MainBlocks, and can be used
   to represent "compacted" relations such as directories, in which few
   records are deleted during its lifetime. The MainFile class is
   better suited for storing true relations.

   The DirectoryFile stores homogeneous records (adhering to the same
   schema) and provides capabilties to:

   - Insert a record in the file (returning the Record Address of the
     added record)
   - Remove a record (given its Record Address) 
   - Retrieve a record (given it's record address)
   - Iterate over all records in the directory file
   
   $Id: DirectoryFile.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db.file;


import java.io.*;
import java.util.*;
import java.lang.ref.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.schema.*;


public final class DirectoryFile
{
    private BufferManager bm;
    private Schema recordSchema;
    private MainBlockFactory recordFactory;

    /* we permanently keep the first block of the linked list of blocks in memory */
    private MainBlock current; 
    private BlockAddress root;
    
    /* we maintain weak references to our spawned iterators so that we
     * can close them when the file is closed */
    private List iterators;

    /* creates a new DirectoryFile */
    public DirectoryFile(BufferManager bm, Schema recordSchema) 
	throws BufferManagerException, DiskManagerException, IOException
    {
	this(bm, recordSchema, null);
    }

    /* loads an existing dir file whose root block can be found
     * at BlockAddress root */
    public DirectoryFile(BufferManager bm, Schema recordSchema, BlockAddress root)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.bm = bm;
	this.recordSchema = recordSchema;
	this.recordFactory = new MainBlockFactory(recordSchema);
	this.iterators = new LinkedList();

	/* load the root block */
	if (root == null) {
	    /* allocate a new root */
	    this.current = (MainBlock) bm.newBlock(recordFactory);
	    this.root = current.getBlockAddress();
	}
	else {
	    this.current = (MainBlock) bm.pinBlock(root, recordFactory);
	    this.root = root;
	}
    }

    /** Destroys the directoryfile and de-allocates all the blocks
     * contained in it */
    public void destroy()
	throws BufferManagerException, DiskManagerException, IOException
    {
	MainBlock current = (MainBlock) bm.pinBlock(root, recordFactory);
	BlockAddress addr = current.getNextBlockAddress();
	while (addr.isValid()) {
	    MainBlock next = (MainBlock) bm.pinBlock(addr, recordFactory);
	    bm.freeBlock(current);
	    current = next;
	    addr = current.getNextBlockAddress();
	}
	close();
    }
    

    /* returns the block address where this file starts */
    public BlockAddress getBlockAddress()
    {
	return root;
    }

    /* Add a record to the file and return the record address of where
     * we added it.*/
    public RecordAddress addRecord(Record rec)
	throws SchemaException, BufferManagerException, DiskManagerException, IOException
    {
	/* we always add in the current block */
	if (current.isFull()) {
	    MainBlock newcurrent = (MainBlock) bm.newBlock(recordFactory);
	    newcurrent.setNextBlockAddress(current.getBlockAddress());
	    current.setPrevBlockAddress(newcurrent.getBlockAddress());
	    bm.unpinBlock(current);
	    current = newcurrent;
	}

	RecordAddress result = null;
	try {
	    int slot = current.addRecord(rec);
	    result =  new RecordAddress(current.getBlockAddress(), slot);
	}
	catch(BlockFullException e) {
	    //cannot happen since we check if it is full
	    e.printStackTrace();
	    System.exit(1);
	}

	return result;
    }

    /* Gets a specific record */
    public Record getRecord(RecordAddress addr)
	throws IndexOutOfBoundsException, BufferManagerException,
	       DiskManagerException, IOException
    {
	BlockAddress baddr = addr.getBlockAddress();
	if (!baddr.isValid())
	    throw new IndexOutOfBoundsException("Block address is not valid.");

	MainBlock block = (MainBlock) bm.pinBlock(baddr, recordFactory);
	Record result = block.getRecord(addr.getRecordIndex());
	bm.unpinBlock(block); //don't forget to unpin it
	return result;
    }

    /* remove a record from the linked list. If the block becomes
     * empty we remove it from the list */
    public void removeRecord(RecordAddress addr)
	throws IndexOutOfBoundsException, BufferManagerException,
	       DiskManagerException, IOException
    {
	BlockAddress baddr = addr.getBlockAddress();
	if (baddr == null || !baddr.isValid())
	    throw new IndexOutOfBoundsException("Block address is not valid.");

	MainBlock block = (MainBlock) bm.pinBlock(baddr, recordFactory);
	block.removeRecord(addr.getRecordIndex());

	/* did the block become empty? If so we remove it (provided it is not the root block) */
	if (block.size() == 0 && !block.getBlockAddress().equals(root)) {

	    BlockAddress prev = block.getPrevBlockAddress();
	    BlockAddress next = block.getNextBlockAddress();

//  	    System.out.println("Block " + block.getBlockAddress().getBlockOffset() + " has become empty and will be freed");
// 	    System.out.println("Prev =" + prev.getBlockOffset());
// 	    System.out.println("Next =" + next.getBlockOffset());
	    
	    MainBlock neighbour;
	    if (prev.isValid()) {
		neighbour = (MainBlock) bm.pinBlock(prev, recordFactory);
		neighbour.setNextBlockAddress(next);
		bm.unpinBlock(neighbour); //don't forget to unpin it
	    }

	    if (next.isValid()) {
		neighbour = (MainBlock) bm.pinBlock(next, recordFactory);
		neighbour.setPrevBlockAddress(prev);
		bm.unpinBlock(neighbour); //don't forget to unpin it
	    }

	    // free the block from which we removed the record
	    bm.freeBlock(block);
	}
	else
	    bm.unpinBlock(block); //don't forget to unpin it
    }

    /* update a record given its record addresss */
    public void updateRecord(RecordAddress addr, Record rec)
	throws IndexOutOfBoundsException, SchemaException,
	       DiskManagerException, BufferManagerException, IOException
    {
	BlockAddress baddr = addr.getBlockAddress();
	if (baddr == null || !baddr.isValid())
	    throw new IndexOutOfBoundsException("Block address is not valid.");

	MainBlock block = (MainBlock) bm.pinBlock(baddr, recordFactory);
	block.updateRecord(addr.getRecordIndex(), rec);
	bm.unpinBlock(block); //don't forget to unpin it
    }
    
    /* make sure all resources are freed when the file is garbage-collected */ 
    public void finalize()
    {
	close();
    }
    
    /* Returns a new iterator. We maintain a list of weak references
     * to spawned iterators such that we can close all iterators when
     * the file is closed. We use WEAK references to allow the spawned
     * iterators to be gargbage-collected. */
    public RecordIterator getIterator() throws BufferManagerException, DiskManagerException, IOException
    {
	RecordIterator result = new DirIterator(this);
	WeakReference ref = new WeakReference(result);
	iterators.add(ref);
	return result;
    }

    /* closes the file - which frees all occupied resources */
    public void close()
    {
	//close all spawned iterators
	ListIterator it = iterators.listIterator();
	while (it.hasNext()) {
	    WeakReference ref = (WeakReference) it.next();
	    RecordIterator rit = (RecordIterator) ref.get();
	    if (rit != null)
		rit.close();
	}
	iterators.clear();

	//unpin the current block
	try {
	    if (current != null)
		bm.unpinBlock(current);
	    current = null;
	    root = null;
	}
	catch(Exception e) {
	    e.printStackTrace();
	    //should not happen
	    System.exit(1);
	}
    }

	    
    /* iterator over all records in the linked list of blocks */
    private final static class DirIterator implements RecordIterator
    {
	private DirectoryFile file;
	private MainBlock current;
	private RecordIterator blockIt; //iterator within a block
	private RecordAddress raddr = null;
	
	public DirIterator(DirectoryFile file) 
	    throws BufferManagerException, DiskManagerException, IOException
	{
	    this.file = file;

	    //only if the file has not been closed
	    if (file.root != null) {
		//pin the first directory block
		current = (MainBlock) file.bm.pinBlock(file.getBlockAddress(), file.recordFactory);
		blockIt = current.getIterator();
	    } 
	}

	public boolean hasNext()
	{
	    //we assume that we do not have empty blocks in the linked list of blocks.
	    return ((blockIt!= null && blockIt.hasNext()) ||
		    (current != null && current.getNextBlockAddress().isValid()));
	}

	public Record next()
	    throws NoSuchElementException, IOException, BufferManagerException, DiskManagerException
	{
	    if (blockIt.hasNext())
		return blockIt.next();

	    BlockAddress next = current.getNextBlockAddress();
	    if (!next.isValid())
		throw new NoSuchElementException();
	    
	    file.bm.unpinBlock(current);
	    current = (MainBlock) file.bm.pinBlock(next, file.recordFactory);
	    blockIt = current.getIterator();
	    return blockIt.next();
	}

	/* closes the attached resources. Calling other methods after
	 * a call to close() can fail */
	public void close()
	{
	    if (blockIt != null)
		blockIt.close();
	    blockIt = null;
	    
	    if (current != null) {
		//unpin the buffer
		try {
		    file.bm.unpinBlock(current);
		}
		catch(Exception e) {
		    //shouldn't happen
		    e.printStackTrace();
		    System.exit(1);
		}

		current = null;
	    }
	    
	    file = null;
	}
	

	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    if (blockIt == null)
		throw new NoSuchElementException();

	    return blockIt.getRecordAddress();
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) 
	    throws  SchemaException, BufferManagerException, DiskManagerException, IOException
	{
	    if (blockIt == null)
		 throw new NoSuchElementException();
	   
	    blockIt.updateRecord(rec);
	}

	/* if the iterator is garbage-collected we should close to free the associated resources */
	public void finalize()
	{
	    close();
	}
	
    }
}

