/*-------------------------------------------------------------------
   Authors: 2001 - 2003: Stijn Vansummeren

   The MainBlock class is a block providing capabilities to:
   - Insert data records in a block
   - Remove data records from a block 
   - Get data records inside a block 

   We assume main blocks are homogeneous: we only store records
   adhering to the same schema.

   Main blocks can be linked in a linked list of main blocks.

   $Id: MainBlock.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
   -------------------------------------------------------------------*/
package be.ulb.db.file;


/*-------------------------------------------------------------------
   Organization of the data inside the block:

   - The first two integers constitue the header and denote the
   previous and next blockaddresses (in a linked list of blocks).
   
   - The next integer specifies the number of records stored in the
     block

   - The remaining space is divided into slots of (1 +
   recordSchema.bytesize()) bytes, where recordSchema is the schema
   describing the records that are stored in the block. Every slot
   stores one record. The extra byte is used to signal if the slot is
   free or not.

  -------------------------------------------------------------------*/


import java.util.*;
import java.io.*;
import java.nio.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.schema.*;

public final class MainBlock extends Block
{
    private static final class HeaderSchema extends Schema
    {
	private static String PREV = "prev";
	private static String NEXT = "next";
	private static String SIZE = "size";
    
	public HeaderSchema()
	{
	    try {
		String [] names =  {PREV, NEXT, SIZE};
		AttributeType [] types = {new AttributeTypeBlockAddress(),new AttributeTypeBlockAddress(),
					  new AttributeTypeInt() };
		
		initSchema(names, types);
	    } catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }
	}

	/* statically determine the offset of the attributes */
	public static int prevIndex = 0;
	public static int nextIndex = 1;
	public static int sizeIndex = 2;
    }

    /* describes the schema of the header */
    protected static HeaderSchema headerSchema = new HeaderSchema();
    
    protected Record header;
    protected Schema recordSchema;
    protected int slotSize;
    protected boolean bIsModified;

    public MainBlock(Schema recordSchema, ByteBuffer buff, BlockAddress addr)
    {
	super(buff, addr);

	/* we start with unmodified contents */
	bIsModified = false;
	
	this.recordSchema = recordSchema;
	this.slotSize = recordSchema.bytesize() + 1;

	/* connect the first Headerschema.headersize bytes of the
	buffer to the header record. We can then write the block's
	header through the record interface */
	buff.rewind();
	buff.limit(headerSchema.bytesize());
	try {
	    header = new Record(headerSchema, buff.slice()); 
	}
	catch (SchemaException e)   {
	    e.printStackTrace();
	    System.exit(1); // it should not occur	  
	}
	
	/* reset the limit to it's full capacity or we won't be able
	 * to write the other records */
	buff.limit(buff.capacity());
    }

    /* initialize a freshly allocated buffer. In particular we give *
     default values for the blocks header, and clear the contents of
     the rest of the block (this makes sure that all slots are
     automatically signaled as unoccupied). In this way "garbage *
     data" from a previous allocation can never be falsly
     interpreted */
    public void initContents()
    {
	setPrevBlockAddress(BlockAddress.INVALIDADDR);
	setNextBlockAddress(BlockAddress.INVALIDADDR);
	try {
	    header.put(HeaderSchema.sizeIndex, new Integer(0));
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}

	/* a newly allocated bytebuffer always has its contents set to 0's */
	ByteBuffer temp = ByteBuffer.allocate(buff.capacity() - headerSchema.bytesize());
	buff.position(headerSchema.bytesize());
	buff.put(temp);

	/* make sure to write these changes back to disk */
	bIsModified = true;
    }

    /* We should return true if the modifications we have made to the
       wrapped buffer has to be saved back to disk */
    public boolean isModified()
    {
	return bIsModified;
    }
    


    public void setPrevBlockAddress(BlockAddress addr)
    {
	try {
	    header.putBlockAddress(HeaderSchema.prevIndex, addr);
	    /* make sure to write these changes back to disk */
	    bIsModified = true;
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
    }

    public void setNextBlockAddress(BlockAddress addr)
    {
	try {
	    header.putBlockAddress(HeaderSchema.nextIndex, addr); 
	    /* make sure to write these changes back to disk */
	    bIsModified = true;
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
    }

    public BlockAddress getPrevBlockAddress()
    {
	BlockAddress result = null;

	try {
	    result =  header.getBlockAddress(HeaderSchema.prevIndex);
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}

	return result;
    }
    
    public BlockAddress getNextBlockAddress()
    {
	BlockAddress result = null;
	
	try {
	    result =  header.getBlockAddress(HeaderSchema.nextIndex);
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur	   
	}

	return result;
    }

    /* capacity is the number of (homogenous) records we can store in a block */
    public int capacity()
    {
	return (buff.capacity() - headerSchema.bytesize())/ slotSize;
    }

    /* returns the number of records currently stored in this block */
    public int size()
    {
	int result = 0;

	try {
	    result = header.getInt(HeaderSchema.sizeIndex).intValue();
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}

	return result;
    }

    /* returns true if the block is full */
    public boolean isFull()
    {
	return capacity() <= size();
    }
    
    /* adds a record in the block. Returns the index where the record
     * is inserted. Throws BlockFullException if there is no room to
     * add the record. Throws SchemaException if the schema of rec
     * does not equal the schema of this block */ 
    public int addRecord(Record rec) throws BlockFullException, SchemaException
    {
	if (!rec.getSchema().equals(recordSchema))
	    throw new SchemaException("Schema of record to be stored differs from the specified schema.");
	
	if (isFull())
	    throw new BlockFullException();

	int i = -1;

	try {

	    boolean found = false;
	    for (i = 0;  i < capacity() && !found; i++) {
		/* check if slot i is free. The first byte in the slot is
		 * 0 if the slot is free */
		if (buff.get(headerSchema.bytesize() + (i * slotSize)) == 0)
		    found = true;
	    }

	    i = i - 1; /* i is one slot too far */
	    buff.position(headerSchema.bytesize() + (i * slotSize));
	    buff.put((byte) 1); // indicate that the slot is occupied
	    buff.put(rec.getByteBuffer());
	    header.put(HeaderSchema.sizeIndex, new Integer(size() + 1));
	}
	catch(Exception e)   {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}
	
	/* make sure to write these changes back to disk */
	bIsModified = true;

	return i;
	
    }

    /* removes record at index slot from this block. We do NOT leave a tombstone. */
    public void removeRecord(int slot) throws IndexOutOfBoundsException
    {
	/* check if the slot is valid */
	if (slot >= capacity() || slot < 0)
	    throw new IndexOutOfBoundsException("Invalid slot specified");

	buff.position(headerSchema.bytesize() + (slot * slotSize));
	byte occupied = buff.get();
	
	if (occupied == 0)
	    throw new IndexOutOfBoundsException("Slot not occupied by a record");

	/* we simply indicate that the slot has become free */
	try {
	    buff.position(headerSchema.bytesize() + (slot * slotSize));
	    buff.put((byte) 0);
	    header.put(HeaderSchema.sizeIndex, new Integer(size() - 1 <= 0 ? 0: size() - 1));
	}
	catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}

	/* make sure to write these changes back to disk */
	bIsModified = true;
    }
    
    /* returns the record at index slot. Throws IndexOutOfBoundsException
     * if the slot is not occupied */
    public Record getRecord(int slot) throws IndexOutOfBoundsException
    {
	Record result = null;

	if (slot >= capacity() || slot < 0)
	    throw new IndexOutOfBoundsException("Invalid slot specified");

	try {
	    buff.position(headerSchema.bytesize() + (slot * slotSize));
	    if (buff.get() == 0)
		throw new IndexOutOfBoundsException("Slot " + slot + " does not contain a record.");
	    
	    buff.limit(headerSchema.bytesize() + ((slot + 1) * slotSize));
	    /* don't share our buffer with the record's
	     * buffer. Records need to be explicitely updated. */
	    ByteBuffer recbuff = ByteBuffer.allocate(recordSchema.bytesize());
	    recbuff.put(buff);
	    result = new Record(recordSchema, recbuff);
	    
	    /* reset the limit to it's full capacity or we won't be able
	     * to write the other records */
	    buff.limit(buff.capacity()); 
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); // shouldn't happen
	}
	
	/* make sure to write these changes back to disk */
	bIsModified = true;

	return result;
    }

    /* updates the record with index slot by record rec */
    public void updateRecord(int slot, Record rec) throws IndexOutOfBoundsException, SchemaException
    {
	if (!rec.getSchema().equals(recordSchema))
	    throw new SchemaException("Schema of record to be stored differs from the specified schema.");

	if (slot >= capacity() || slot < 0)
	    throw new IndexOutOfBoundsException("Invalid slot specified");

	try {
	    buff.position(headerSchema.bytesize() + (slot * slotSize));
	    if (buff.get() == 0)
		throw new IndexOutOfBoundsException("Slot " + slot + " does not contain a record.");

	    buff.put(rec.getByteBuffer());
	}
	catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1); // it should not occur
	}

	/* make sure to write these changes back to disk */
	bIsModified = true;

    }
    
    /* returns an object capable of iterating over all the records in the block */
    public RecordIterator getIterator()
    {
	return new MainRecordIterator(this);
    }
    

    /* The iterator for the MainBlock */
    protected final static class MainRecordIterator implements RecordIterator
    {
	private MainBlock block;
	private int currentSlot;
	private boolean bhasCurrent;
	private int nextSlot;
	private boolean bhasNext;

	public MainRecordIterator(MainBlock block)
	{
	    this(block, 0);
	}

	/* create an iterator on block, starting from startSlot */
	public MainRecordIterator(MainBlock block, int startSlot)
	{
	    this.block = block;
	    
	    currentSlot = startSlot - 1;
	    bhasCurrent = false;
	    findNext();
	}

	
	public boolean hasNext()
	{
	    return bhasNext;
	}

	public Record next() throws NoSuchElementException
	{
	    Record result = null;

	    if (!bhasNext)
		throw new NoSuchElementException();
	    
	    try {
		currentSlot = nextSlot;
		result = block.getRecord(currentSlot);
		findNext();
		bhasCurrent = true;
	    }
	    catch(IndexOutOfBoundsException e) {
		e.printStackTrace();
		System.exit(1); // it should not occur
	    }

	    return result;
	}

	protected void findNext() 
	{
	    int i;
	    boolean found = false;

	    for (i = currentSlot + 1;  i < block.capacity() && !found; i++) {
		/* check if slot i is occupied. The first byte in the slot is
		 * 0 if the slot is free */
		if (block.buff.get(headerSchema.bytesize() + (i * block.slotSize)) != 0)
		    found = true;
	    }

	    bhasNext = found;
	    nextSlot = i - 1;
	}

	/* closes the attached resources. Calling other methods after a call to close() can fail */
	public void close()
	{
	    bhasCurrent = bhasNext = false;
	}
	
	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    if (!bhasCurrent)
		throw new NoSuchElementException();
	    
	    return new RecordAddress(block.getBlockAddress(), currentSlot);
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) throws NoSuchElementException, SchemaException
	{
	    if (!bhasCurrent)
		throw new NoSuchElementException();
	    
	    try {
		block.updateRecord(currentSlot, rec);
	    }
	    catch(IndexOutOfBoundsException e) {
		throw new NoSuchElementException(e.getMessage());
	    }
	}
    }
}
