/*-------------------------------------------------------------------
   Authors: 2001 - 2003: Stijn Vansummeren

   The MainFile is a collection of MainBlocks, and can represent the
   "data" records in a relation. It provides capabilities to:

   - Insert a main record in the main file of a relation (returning
     the Record Address of the added record)

   - Remove a main record (given its Record Address) 
   - Retrieve a record given it's record address 
   - Iterate over all records in the main file
   
   $Id: MainFile.java,v 1.2 2006-12-04 13:11:15 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db.file;


import java.io.*;
import java.lang.*;
import java.util.*;
import java.lang.ref.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.schema.*;


/*-------------------------------------------------------------------
  We maintain a directory holding information of the data blocks
  constituting the MainFile. Data blocks are of type MainBlock when
  loaded into memory.  The directory itself is also composed of
  MainBlocks, with each record having schema DataBlockInfoSchema as
  defined below.

  The directory is implemented by the DirectoryFile classs (All
  directory blocks are in a doubly-linked list of blocks). Each
  directory entry points to a single data block, which contains the
  actual records.

  -------------------------------------------------------------------*/

public final class MainFile
{
    /* Describes the schema of the DataBlockInfo records */
    private static final class DataBlockInfoSchema extends Schema
    {
	private static String ADDR = "addr";
	private static String FULL = "full";

	public DataBlockInfoSchema()
	{
	    try {
		String [] names =  {ADDR, FULL};
		AttributeType [] types = {new AttributeTypeBlockAddress(), new AttributeTypeBoolean() };
		
		initSchema(names, types);
	    } catch(SchemaException e) {
		e.printStackTrace();
		System.exit(1);
		//should not occur
	    }
	}

	/* statically determine the offset of the attributes */
	public static int addrIndex = 0;
	public static int fullIndex = 1;
    }

    private static Schema dataBlockInfoSchema = new DataBlockInfoSchema();

    private BufferManager bm;
    private Schema recordSchema;
    private MainBlockFactory recordFactory;
    private DirectoryFile directory;

    private RecordAddress addCandidate; //the directory record address pointing to a data block where there is some free space.

    /* we maintain weak references to our spawned iterators so that we
     * can close them when the file is closed */
    private List iterators;

    /* creates a new mainfile */
    public MainFile(BufferManager bm, Schema recordSchema) 
	throws BufferManagerException, DiskManagerException, IOException
    {
	this(bm, recordSchema, null);
    }

    /* loads an existing mainfile whose root block (containing the
     * first block of the directory file) can be found at BlockAddress
     * root */
    public MainFile(BufferManager bm, Schema recordSchema, BlockAddress root)
	throws BufferManagerException, DiskManagerException, IOException
    {
	this.bm = bm;
	this.recordSchema = recordSchema;
	this.recordFactory = new MainBlockFactory(recordSchema);
	this.iterators = new LinkedList();

	if (root == null) 
	    /* create a new directory file */
	    directory = new DirectoryFile(bm, dataBlockInfoSchema);
	else
	    /* load directory file  */
	    directory = new DirectoryFile(bm, dataBlockInfoSchema, root);
    }

    /*Destroys the mainfile and de-allocates all the blocks contained
     * in it */
    public void destroy()
	throws BufferManagerException, DiskManagerException, IOException
    {
	try {
	    RecordIterator it = directory.getIterator();
	    while (it.hasNext()) {
		Record rec = it.next();
		BlockAddress addr = rec.getBlockAddress(DataBlockInfoSchema.addrIndex);

		//TODO: should be possible without pinning a block first
		Block block = bm.pinBlock(addr, recordFactory);
		bm.freeBlock(block);
	    }
	    
	    directory.destroy();

	    close();
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); //should not occur
	}
    }
    

    /* returns the block address from where the MainFile can be loaded next time */
    public BlockAddress getBlockAddress()
    {
	return directory.getBlockAddress();
    }
    
    /* Add a main record to a main file and return the record address
     * of where we added it.*/
    public RecordAddress addRecord(Record record)
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordAddress result = null;
	
	try {
	    /* find a data block with some free space */
	    Record dirrec = null;
	    boolean found = false;

	    //first see if candidateAddr points to an info record
	    //whose data block has some free space
	    if (addCandidate != null ) {
		try {
		    dirrec = directory.getRecord(addCandidate);
		    found = !dirrec.getBoolean(DataBlockInfoSchema.fullIndex).booleanValue();
		}
		catch(IndexOutOfBoundsException e) {
		    //ignore - can occur if the candidate directory
		    //record has been deleted.
		}
	    }

	    //otherwise find another candidate
	    if (!found) {
		RecordIterator it = directory.getIterator();

		while (!found && it.hasNext()) {
		    dirrec = it.next();
		    found = !dirrec.getBoolean(DataBlockInfoSchema.fullIndex).booleanValue() ;
		}
		
		if (found) 
		    addCandidate = it.getRecordAddress();
		
		/* free resources */
		it.close();
	    }
	    
	    MainBlock datablock;

	    /* do we now really have a data block we can insert a
	     * record into ? */
	    if (found) {
		datablock = (MainBlock) bm.pinBlock(dirrec.getBlockAddress(DataBlockInfoSchema.addrIndex), recordFactory);
	    }
	    else {
		/* create a new data block and put it in the directory */
		datablock = (MainBlock) bm.newBlock(recordFactory);

		dirrec = new Record(dataBlockInfoSchema);
		dirrec.putBlockAddress(DataBlockInfoSchema.addrIndex, datablock.getBlockAddress());
	    
		addCandidate = directory.addRecord(dirrec);
	    }

	    /* add it to the data block */
	    int slot = datablock.addRecord(record);
	    result =  new RecordAddress(datablock.getBlockAddress(), slot);

	    /* update the directory */
	    dirrec.putBoolean(DataBlockInfoSchema.fullIndex, new Boolean(datablock.isFull()));
	    directory.updateRecord(addCandidate, dirrec);

	    /* free resources */
	    bm.unpinBlock(datablock);
	}
	catch(SchemaException e) {
	    e.printStackTrace();
	    System.exit(1); //should not happen
	}
	catch(BlockFullException e) {
	    e.printStackTrace();
	    System.exit(1); //should not happen unless the directory is corrupt
	}

	return result;
    }

    public Record getRecord(RecordAddress addr)
	throws IndexOutOfBoundsException, BufferManagerException,
	       DiskManagerException, IOException
    {
	BlockAddress baddr = addr.getBlockAddress();
	if (!baddr.isValid())
	    throw new IndexOutOfBoundsException("Block address is not valid.");

	MainBlock block = (MainBlock) bm.pinBlock(baddr, recordFactory);
	Record result = block.getRecord(addr.getRecordIndex());
	bm.unpinBlock(block); //don't forget to unpin it
	return result;
    }

    public void removeRecord(RecordAddress addr)
    throws IndexOutOfBoundsException, BufferManagerException,
    DiskManagerException, IOException
    {
    	try {

    		BlockAddress baddr = addr.getBlockAddress();
    		if (!baddr.isValid())
    			throw new IndexOutOfBoundsException("Block address is not valid.");

    		MainBlock datablock = (MainBlock) bm.pinBlock(baddr, recordFactory);

    		datablock.removeRecord(addr.getRecordIndex());
    		//has the datablock become empty?
    		Record dirrec = null;
    		RecordAddress diraddr = null;

    		RecordIterator it = directory.getIterator();
    		boolean found = false;
    		while (!found && it.hasNext()) {
    			dirrec = it.next();
    			found = dirrec.getBlockAddress(DataBlockInfoSchema.addrIndex).equals(datablock.getBlockAddress());
    		}

    		if (!found)
    			throw new IndexOutOfBoundsException("Data block cannot be found in MainFile Directory.");

    		diraddr = it.getRecordAddress();

    		if (datablock.size() == 0) {
    			directory.removeRecord(diraddr);
    			bm.freeBlock(datablock);

    			//if we just deleted our candidate - set it to null so
    			//that we search for a new candidate when a record is
    			//added
    			if (it.getRecordAddress().equals(addCandidate)) 
    				addCandidate = null;
    		}
    		else {
    			//we just deleted a record - the datablock can never be full!
    			dirrec.putBoolean(DataBlockInfoSchema.fullIndex, new Boolean(false));
    			directory.updateRecord(diraddr, dirrec);
    			bm.unpinBlock(datablock); 

    			addCandidate = it.getRecordAddress();
    		}

    		//free resources
    		it.close();
    	}
    	catch(SchemaException e) {
    		e.printStackTrace();
    		System.exit(1); //should not happen
    	}
    }

    public void updateRecord(RecordAddress addr, Record record)
	throws IndexOutOfBoundsException, SchemaException, BufferManagerException,
	       DiskManagerException, IOException
    {
	BlockAddress baddr = addr.getBlockAddress();
	if (!baddr.isValid())
	    throw new IndexOutOfBoundsException("Block address is not valid.");

	MainBlock datablock = (MainBlock) bm.pinBlock(baddr, recordFactory);
	datablock.updateRecord(addr.getRecordIndex(), record);
	bm.unpinBlock(datablock); //don't forget to unpin it
    }

    /* make sure all resources are freed when the file is garbage-collected */ 
    public void finalize()
    {
	close();
    }

    /* Returns a new iterator. We maintain a list of weak references
     * to spawned iterators such that we can close all iterators when
     * the file is closed. We use WEAK references to allow the spawned
     * iterators to be gargbage-collected. */
    public RecordIterator getIterator()
	throws BufferManagerException, DiskManagerException, IOException
    {
	RecordIterator result = new DataIterator(this);
	WeakReference ref = new WeakReference(result);
	iterators.add(ref);
	return result;
     }

    /* closes the file - which frees all occupied resources */
    public void close()
    {
	//close all spawned iterators
	ListIterator it = iterators.listIterator();
	while (it.hasNext()) {
	    WeakReference ref = (WeakReference) it.next();
	    RecordIterator rit = (RecordIterator) ref.get();
	    if (rit != null)
		rit.close();
	}
	iterators.clear();

	//close the directory
	if (directory != null)
	    directory.close();

	/* calling methods after closure WILL result in errors */
	directory = null;
	bm = null;
	recordFactory = null;
	recordSchema = null;
    }

    /* iterator over the data records */
    private final static class DataIterator implements RecordIterator
    {
	private MainFile file;
	private MainBlock current;
	private RecordIterator dirIt; //iterator over the directory
	private RecordIterator blockIt; //iterator within a block
	
	public DataIterator(MainFile file)
	    throws BufferManagerException, DiskManagerException, IOException
	{
	    this.file = file;

	    dirIt = file.directory.getIterator();
	    blockIt = null;
	    current = null;
	}

	public boolean hasNext()
	{
	    //we assume non-empty directory blocks
	    return ((blockIt != null && blockIt.hasNext()) || (dirIt != null && dirIt.hasNext()));
	}

	public Record next()
	    throws NoSuchElementException, IOException, BufferManagerException, DiskManagerException
	{
	    if (!hasNext())
		throw new NoSuchElementException();

	    if (blockIt != null && blockIt.hasNext())
		return blockIt.next();

	    Record dirrec = dirIt.next();

	    Record result = null;
	    try {
		if (current != null)
		    file.bm.unpinBlock(current);
		current = (MainBlock) file.bm.pinBlock(dirrec.getBlockAddress(DataBlockInfoSchema.addrIndex), file.recordFactory);
		blockIt = current.getIterator();
		result = blockIt.next();
	    }
	    catch (SchemaException e) {
		e.printStackTrace();
		System.exit(1); //should not occur
	    }

	    return result;
	}

	/* returns the RecordAddress of the current record. Throws
	 * NoSuchElementException if there is no current record*/
	public RecordAddress getRecordAddress() throws NoSuchElementException
	{
	    if (blockIt == null)
		throw new NoSuchElementException();
	    
	    return blockIt.getRecordAddress();
	}
	
	/* updates the current record. Throws NoSuchElementException if
	 * there is no current record. Throws SchemaException if the
	 * schema of rec differs from the record to be updated. */
	public void updateRecord(Record rec) 
	    throws  SchemaException, BufferManagerException, DiskManagerException, IOException
	{
	    if (blockIt == null)
		throw new NoSuchElementException();

	    blockIt.updateRecord(rec);
	}

	/* closes the attached resources. Calling other methods after
	 * a call to close() can fail */
	public void close()
	{
	    if (dirIt != null)
		dirIt.close();
	    dirIt = null;

	    if (blockIt != null)
		blockIt.close();
	    blockIt = null;
	    
	    if (current != null) {
		//unpin the buffer
		try {
		    file.bm.unpinBlock(current);
		}
		catch(Exception e) {
		    //shouldn't happen
		    e.printStackTrace();
		    System.exit(1);
		}

		current = null;
	    }

	    file = null;
	}
	


	/* if the iterator is garbage-collected we should close to free the associated resources */
	public void finalize()
	{
	    close();
	}
    }
    
}

