/*-------------------------------------------------------------------
   Authors: 2003: Stijn Vansummeren

   An index is a data structure which can insert, find and remove
   addresses to a record with some (index) key value.

   $Id: Index.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db.index;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.schema.*;

public abstract class Index
{
    protected Schema keySchema;   //defines the schema of keyvalues
    protected BufferManager bm;
    
    /** Subclasses should create a new index */
    public Index(Schema keySchema, BufferManager bm)
    {
	this.keySchema = keySchema;
	this.bm = bm;
    }

    /** Subclasses should load the index which starts at address addr */
    public Index(Schema keySchema, BufferManager bm, BlockAddress addr)
    {
	this.keySchema = keySchema;
	this.bm = bm;
    }

    /** destroy the index - free all blocks belonging to the index */
    public abstract void destroy();

    /* Close the index. Subclasses should close all resources. (i.e. Unpin
     * all buffers and close all open iterators) */ 
    public abstract void close();

    /** returns the blockaddress from which the index can be loaded next time */
    public abstract BlockAddress getBlockAddress();

    /* returns the schema of the search keys */
    public Schema getSchema()
    {
	return keySchema;
    }

    /** Subclasses should add RecordAddress addr to the index, knowing
     * the record at that address equals keyvalue on its attributes in
     * keySchema.  */
    public void addRecord(Record keyvalue, RecordAddress addr)
	throws SchemaException 
    {
	if (!keyvalue.getSchema().equals(keySchema))
	    throw new SchemaException("The schema of the key does not correspond with the given schema.");
    }
    
    /** Subclasses should search in the index for RecordAddresses
     *  whose corresponding records equal keyvalue in their keySchema
     *  attributes. */
    public RecordAddressIterator findRecord(Record keyvalue)
	throws SchemaException 	
    {
	if (!keyvalue.getSchema().equals(keySchema))
	    throw new SchemaException("The schema of the key does not correspond with the given schema.");

	return null;
    }
    
    /** Subclasses should delete the (key,addr) pair (keyvalue, addr)
     * from the index */
    public void removeRecord(Record keyvalue, RecordAddress addr)
	throws SchemaException 
    {
	if (!keyvalue.getSchema().equals(keySchema))
	    throw new SchemaException("The schema of the key does not correspond with the given schema.");
    }

    public boolean equals(Object o) 
    {
	if (!(o instanceof Index))
	    return false;
	
	Index other = (Index) o;
	return other.keySchema.equals(keySchema);
    }

    public int hashCode() 
    {
	return keySchema.hashCode();
    }


    /* all indexes clean up their resources when garbage-collected */
    public void finalize()
    {
	close();
    }

}
