/*-------------------------------------------------------------------
   Authors: 2003: Stijn Vansummeren
   
   AttributeTypeString represents the field type "CHAR(n)" that can
   occur in a schema. Note that we always reserve n characters for an
   attribute with this type, hence it is NOT varying-length.

   $Id: AttributeTypeString.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
  -------------------------------------------------------------------*/
package be.ulb.db.schema;

import java.nio.*;

public class AttributeTypeString extends AttributeType
{
    private static int MAXLENGTH = 255;
    private int length;

    /* A String field type needs to mention its maximum length (in
     * characters) */
    public AttributeTypeString(int length)
    {
	this.length = length;

	/* prevent maximum values which would require us to span
	 * records across blocks */
	if (length > MAXLENGTH) 
	    throw new IllegalArgumentException("Maximum string length is " + MAXLENGTH);
    }

    /* A java object is a valid instance of this field type if it is a
     * String */
    public boolean hasInstance(Object o)
    {
	return (o instanceof String);
    }

    /* returns the size of a value in bytes */
    public int bytesize()
    {
	//we need two bytes per unicode character and one additional
	//byte to store the true length
	return (length * 2) + 1; 
    }

    /* returns the maximum length of the string */
    public int getLength()
    {
	return length;
    }

    /* writes an instance of the field type to the byte buffer. We
     * assume the bytebuffer's position and limit have been set
     * correctly to the start end end of this field */
    public void writeInstanceToByteBuffer(ByteBuffer buff, Object instance)
    {
	String s = (String) instance;
	byte l = (byte) Math.min(s.length(), length);
	buff.put(l);
	CharBuffer cbuff = buff.asCharBuffer();

	/* The length is in the number of CHARACTERS that should be
	   written (not bytes) */
	cbuff.put(s.toCharArray(), 0, l);

	/* padd the rest with zeroes */ 
	buff.position(buff.position() + (l*2));
	ByteBuffer padding = ByteBuffer.allocate(buff.limit() - buff.position());
	buff.put(padding);
    }
    
    /* reads an instance of the field type from a byte buffer. We
     * assume the bytebuffer's position and limit have been set
     * correctly to the start end end of this field */
    public Object readInstanceFromByteBuffer(ByteBuffer buff)
    {
	byte l = buff.get();
	CharBuffer cbuff = buff.asCharBuffer();
	//	System.out.println("reading string, l = "+ l);
	
	
	//cbuff.limit() should be one behind our stop position and
	//limit gets decremented one last time in the for loop
	cbuff.limit(l);
	return cbuff.toString();
    }
    
    public boolean equals(Object o)
    {
	if (! (o instanceof AttributeTypeString))
	    return false;
	
	AttributeTypeString other = (AttributeTypeString) o;
	return other.length == length;
    }
}
