/*-------------------------------------------------------------------
   Authors: 2003: Stijn Vansummeren
   
   The Schema class provides capabilities to describe a relation
   schema. 

   $Id: Schema.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
-------------------------------------------------------------------*/
package be.ulb.db.schema;

import java.util.*;

public class Schema
{
    protected Hashtable name_to_index;
    protected String [] index_to_name;
    protected AttributeType [] types;
    protected int bytesize = 0;  /* number of bytes needed to store a record adhering to this schema */
    protected int [] offsets;

    /* A schema consists of a sequence of (name, type) tuples.  We can
       init a schema by giving the attribute dnames and corresponding
       types */
    public Schema(String [] names, AttributeType [] types) throws SchemaException
    {
	initSchema(names, types);
    }
    
    /* The schema initialisation code which subclasses can
     * override. We bassically do some sanity-checks and set the
     * internal datastructures */
    protected void initSchema(String [] names, AttributeType [] types) throws SchemaException
    {
	if (names.length != types.length )
	    throw new SchemaException ("Names and types parameters should be of equal length!");
	
	this.types = new AttributeType[types.length];
	this.name_to_index = new Hashtable();
	this.index_to_name = new String[types.length];
	this.offsets = new int[types.length];
	
	for (int i = 0; i < types.length; i++)
	    {
		this.types[i] = types[i];
		offsets[i] = bytesize;
		bytesize += types[i].bytesize();
		
		/* we do not allow a null-name */
		if (names[i] == null)
		    throw new SchemaException("Attribute names cannot be null!");

		/* store it in the name_to_index mapping to allow name resolving */
		if (this.name_to_index.put(names[i], new Integer(i)) != null) 
		    throw new SchemaException("Attribute names should be unique!");
		
		/* also store the name in index_to_name to allow
		 * reverse-lookup */
		this.index_to_name[i] = names[i];
	    }
    }
    
    /* Default constructor is only available to subclasses */
    protected Schema()
    {
    }
    
    /* returns the index of the attribute with attribute name name */
    public int resolveName(String name) throws SchemaException
    {
	Integer attrindex = (Integer) name_to_index.get(name);
	if (attrindex == null) 
	    throw new SchemaException("Attribute with name " + name + " not found");
	
	return attrindex.intValue();
    }

    public String getName(int index) throws SchemaException
    {
	try {
	    return index_to_name[index];
	}
	catch(ArrayIndexOutOfBoundsException e)  {
	    throw new SchemaException("No attribute with index " + index);
	}
    }

    /* returns the type of attribute with attributename name */
    public AttributeType getType(String name) throws SchemaException
    {
	return getType((Integer) name_to_index.get(name));
    }

    /* returns the type of attribute with attributeindex index */
    public AttributeType getType(Integer index) throws SchemaException
    {
	if (index != null)
	    return getType(index.intValue());
	else
	    throw new SchemaException("No attribute with index null");
    }
    
    /* returns the type of attribute with attributeindex index */
    public AttributeType getType(int index) throws SchemaException
    {
	try {
	    return types[index];
	}
	catch(ArrayIndexOutOfBoundsException e)  {
	    throw new SchemaException("No attribute with index " + index);
	}
    }

    /* returns the number of bytes needed to store a record adhering
     * to this schema */
    public int bytesize()
    {
	return bytesize;
    }

    /* returns the offset (in bytes) the attribute at index starts at
     * in a record */
    public int getOffset(int index)
    {
	return offsets[index];
    }

    /* returns the number of attributes in this schema */
    public int length()
    {
	return types.length;
    }

    /* allows to see if the current schema is a subschema of schema other */
    public boolean isSubSchemaOf(Schema other)
    {
	if (this == other)
	    return true;
	
	if (other.length() < this.length())
	    return false;

	boolean subschema = true;
	for (int i = 0; i < types.length && subschema; i++) {
	    Integer otherindex = (Integer) other.name_to_index.get(index_to_name[i]);
	    
	    if (otherindex == null) 
		subschema &= false;
	    else 
		subschema &= types[i].equals(other.types[otherindex.intValue()]);
	}
	
	return subschema;
    }

    /* project the schema on some attriutes */
    public Schema project(String [] names) throws SchemaException
    {
	AttributeType [] types = new AttributeType[names.length];
	for (int i = 0; i < names.length; i++) {
	   try {
	       types[i] = this.types[((Integer) name_to_index.get(names[i])).intValue()];
	   }
	   catch (NullPointerException e) {
	       throw new SchemaException("Scheme does not contain attribute " + names[i]);
	   }
	}
	
	return new Schema(names,types);
    }

    /* allows to see if two schema objects are semantically equal */
    public boolean equals(Object o)
    {
	if (this == o)
	    return true;

	Schema other = (Schema) o;

	if (other.length() != this.length())
	    return false;

	boolean equal = true;
	for (int i = 0; i < length() && equal; i++) {
	    equal &= types[i].equals(other.types[i]);
	    equal &= index_to_name[i].equals(other.index_to_name[i]);
	}
	
	return equal;
    }

    public int hashCode()
    {
	int code = 0;
	
	for (int i = 0; i < index_to_name.length; i++) 
	    code += index_to_name[i].hashCode();
	
	return code / index_to_name.length;
    }
}

